# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
# pylint: disable=protected-access, arguments-differ, signature-differs, broad-except

import calendar
import decimal
import functools
import sys
import logging
import base64
import re
import copy
import typing
import enum
import email.utils
from datetime import datetime, date, time, timedelta, timezone
from json import JSONEncoder
from typing_extensions import Self
import isodate
from azure.core.exceptions import DeserializationError
from azure.core import CaseInsensitiveEnumMeta
from azure.core.pipeline import PipelineResponse
from azure.core.serialization import _Null

if sys.version_info >= (3, 9):
    from collections.abc import MutableMapping
else:
    from typing import MutableMapping

_LOGGER = logging.getLogger(__name__)

__all__ = ["SdkJSONEncoder", "Model", "rest_field", "rest_discriminator"]

TZ_UTC = timezone.utc
_T = typing.TypeVar("_T")


def _timedelta_as_isostr(td: timedelta) -> str:
    """Converts a datetime.timedelta object into an ISO 8601 formatted string, e.g. 'P4DT12H30M05S'

    Function adapted from the Tin Can Python project: https://github.com/RusticiSoftware/TinCanPython

    :param timedelta td: The timedelta to convert
    :rtype: str
    :return: ISO8601 version of this timedelta
    """

    # Split seconds to larger units
    seconds = td.total_seconds()
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    days, hours = divmod(hours, 24)

    days, hours, minutes = list(map(int, (days, hours, minutes)))
    seconds = round(seconds, 6)

    # Build date
    date_str = ""
    if days:
        date_str = "%sD" % days

    if hours or minutes or seconds:
        # Build time
        time_str = "T"

        # Hours
        bigger_exists = date_str or hours
        if bigger_exists:
            time_str += "{:02}H".format(hours)

        # Minutes
        bigger_exists = bigger_exists or minutes
        if bigger_exists:
            time_str += "{:02}M".format(minutes)

        # Seconds
        try:
            if seconds.is_integer():
                seconds_string = "{:02}".format(int(seconds))
            else:
                # 9 chars long w/ leading 0, 6 digits after decimal
                seconds_string = "%09.6f" % seconds
                # Remove trailing zeros
                seconds_string = seconds_string.rstrip("0")
        except AttributeError:  # int.is_integer() raises
            seconds_string = "{:02}".format(seconds)

        time_str += "{}S".format(seconds_string)
    else:
        time_str = ""

    return "P" + date_str + time_str


def _serialize_bytes(o, format: typing.Optional[str] = None) -> str:
    encoded = base64.b64encode(o).decode()
    if format == "base64url":
        return encoded.strip("=").replace("+", "-").replace("/", "_")
    return encoded


def _serialize_datetime(o, format: typing.Optional[str] = None):
    if hasattr(o, "year") and hasattr(o, "hour"):
        if format == "rfc7231":
            return email.utils.format_datetime(o, usegmt=True)
        if format == "unix-timestamp":
            return int(calendar.timegm(o.utctimetuple()))

        # astimezone() fails for naive times in Python 2.7, so make make sure o is aware (tzinfo is set)
        if not o.tzinfo:
            iso_formatted = o.replace(tzinfo=TZ_UTC).isoformat()
        else:
            iso_formatted = o.astimezone(TZ_UTC).isoformat()
        # Replace the trailing "+00:00" UTC offset with "Z" (RFC 3339: https://www.ietf.org/rfc/rfc3339.txt)
        return iso_formatted.replace("+00:00", "Z")
    # Next try datetime.date or datetime.time
    return o.isoformat()


def _is_readonly(p):
    try:
        return p._visibility == ["read"]  # pylint: disable=protected-access
    except AttributeError:
        return False


class SdkJSONEncoder(JSONEncoder):
    """A JSON encoder that's capable of serializing datetime objects and bytes."""

    def __init__(self, *args, exclude_readonly: bool = False, format: typing.Optional[str] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.exclude_readonly = exclude_readonly
        self.format = format

    def default(self, o):  # pylint: disable=too-many-return-statements
        if _is_model(o):
            if self.exclude_readonly:
                readonly_props = [p._rest_name for p in o._attr_to_rest_field.values() if _is_readonly(p)]
                return {k: v for k, v in o.items() if k not in readonly_props}
            return dict(o.items())
        try:
            return super(SdkJSONEncoder, self).default(o)
        except TypeError:
            if isinstance(o, _Null):
                return None
            if isinstance(o, decimal.Decimal):
                return float(o)
            if isinstance(o, (bytes, bytearray)):
                return _serialize_bytes(o, self.format)
            try:
                # First try datetime.datetime
                return _serialize_datetime(o, self.format)
            except AttributeError:
                pass
            # Last, try datetime.timedelta
            try:
                return _timedelta_as_isostr(o)
            except AttributeError:
                # This will be raised when it hits value.total_seconds in the method above
                pass
            return super(SdkJSONEncoder, self).default(o)


_VALID_DATE = re.compile(r"\d{4}[-]\d{2}[-]\d{2}T\d{2}:\d{2}:\d{2}" + r"\.?\d*Z?[-+]?[\d{2}]?:?[\d{2}]?")
_VALID_RFC7231 = re.compile(
    r"(Mon|Tue|Wed|Thu|Fri|Sat|Sun),\s\d{2}\s"
    r"(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s\d{4}\s\d{2}:\d{2}:\d{2}\sGMT"
)


def _deserialize_datetime(attr: typing.Union[str, datetime]) -> datetime:
    """Deserialize ISO-8601 formatted string into Datetime object.

    :param str attr: response string to be deserialized.
    :rtype: ~datetime.datetime
    :returns: The datetime object from that input
    """
    if isinstance(attr, datetime):
        # i'm already deserialized
        return attr
    attr = attr.upper()
    match = _VALID_DATE.match(attr)
    if not match:
        raise ValueError("Invalid datetime string: " + attr)

    check_decimal = attr.split(".")
    if len(check_decimal) > 1:
        decimal_str = ""
        for digit in check_decimal[1]:
            if digit.isdigit():
                decimal_str += digit
            else:
                break
        if len(decimal_str) > 6:
            attr = attr.replace(decimal_str, decimal_str[0:6])

    date_obj = isodate.parse_datetime(attr)
    test_utc = date_obj.utctimetuple()
    if test_utc.tm_year > 9999 or test_utc.tm_year < 1:
        raise OverflowError("Hit max or min date")
    return date_obj


def _deserialize_datetime_rfc7231(attr: typing.Union[str, datetime]) -> datetime:
    """Deserialize RFC7231 formatted string into Datetime object.

    :param str attr: response string to be deserialized.
    :rtype: ~datetime.datetime
    :returns: The datetime object from that input
    """
    if isinstance(attr, datetime):
        # i'm already deserialized
        return attr
    match = _VALID_RFC7231.match(attr)
    if not match:
        raise ValueError("Invalid datetime string: " + attr)

    return email.utils.parsedate_to_datetime(attr)


def _deserialize_datetime_unix_timestamp(attr: typing.Union[float, datetime]) -> datetime:
    """Deserialize unix timestamp into Datetime object.

    :param str attr: response string to be deserialized.
    :rtype: ~datetime.datetime
    :returns: The datetime object from that input
    """
    if isinstance(attr, datetime):
        # i'm already deserialized
        return attr
    return datetime.fromtimestamp(attr, TZ_UTC)


def _deserialize_date(attr: typing.Union[str, date]) -> date:
    """Deserialize ISO-8601 formatted string into Date object.
    :param str attr: response string to be deserialized.
    :rtype: date
    :returns: The date object from that input
    """
    # This must NOT use defaultmonth/defaultday. Using None ensure this raises an exception.
    if isinstance(attr, date):
        return attr
    return isodate.parse_date(attr, defaultmonth=None, defaultday=None)  # type: ignore


def _deserialize_time(attr: typing.Union[str, time]) -> time:
    """Deserialize ISO-8601 formatted string into time object.

    :param str attr: response string to be deserialized.
    :rtype: datetime.time
    :returns: The time object from that input
    """
    if isinstance(attr, time):
        return attr
    return isodate.parse_time(attr)


def _deserialize_bytes(attr):
    if isinstance(attr, (bytes, bytearray)):
        return attr
    return bytes(base64.b64decode(attr))


def _deserialize_bytes_base64(attr):
    if isinstance(attr, (bytes, bytearray)):
        return attr
    padding = "=" * (3 - (len(attr) + 3) % 4)  # type: ignore
    attr = attr + padding  # type: ignore
    encoded = attr.replace("-", "+").replace("_", "/")
    return bytes(base64.b64decode(encoded))


def _deserialize_duration(attr):
    if isinstance(attr, timedelta):
        return attr
    return isodate.parse_duration(attr)


def _deserialize_decimal(attr):
    if isinstance(attr, decimal.Decimal):
        return attr
    return decimal.Decimal(str(attr))


_DESERIALIZE_MAPPING = {
    datetime: _deserialize_datetime,
    date: _deserialize_date,
    time: _deserialize_time,
    bytes: _deserialize_bytes,
    bytearray: _deserialize_bytes,
    timedelta: _deserialize_duration,
    typing.Any: lambda x: x,
    decimal.Decimal: _deserialize_decimal,
}

_DESERIALIZE_MAPPING_WITHFORMAT = {
    "rfc3339": _deserialize_datetime,
    "rfc7231": _deserialize_datetime_rfc7231,
    "unix-timestamp": _deserialize_datetime_unix_timestamp,
    "base64": _deserialize_bytes,
    "base64url": _deserialize_bytes_base64,
}


def get_deserializer(annotation: typing.Any, rf: typing.Optional["_RestField"] = None):
    if rf and rf._format:
        return _DESERIALIZE_MAPPING_WITHFORMAT.get(rf._format)
    return _DESERIALIZE_MAPPING.get(annotation)


def _get_type_alias_type(module_name: str, alias_name: str):
    types = {
        k: v
        for k, v in sys.modules[module_name].__dict__.items()
        if isinstance(v, typing._GenericAlias)  # type: ignore
    }
    if alias_name not in types:
        return alias_name
    return types[alias_name]


def _get_model(module_name: str, model_name: str):
    models = {k: v for k, v in sys.modules[module_name].__dict__.items() if isinstance(v, type)}
    module_end = module_name.rsplit(".", 1)[0]
    models.update({k: v for k, v in sys.modules[module_end].__dict__.items() if isinstance(v, type)})
    if isinstance(model_name, str):
        model_name = model_name.split(".")[-1]
    if model_name not in models:
        return model_name
    return models[model_name]


_UNSET = object()


class _MyMutableMapping(MutableMapping[str, typing.Any]):  # pylint: disable=unsubscriptable-object
    def __init__(self, data: typing.Dict[str, typing.Any]) -> None:
        self._data = copy.deepcopy(data)

    def __contains__(self, key: typing.Any) -> bool:
        return key in self._data

    def __getitem__(self, key: str) -> typing.Any:
        return self._data.__getitem__(key)

    def __setitem__(self, key: str, value: typing.Any) -> None:
        self._data.__setitem__(key, value)

    def __delitem__(self, key: str) -> None:
        self._data.__delitem__(key)

    def __iter__(self) -> typing.Iterator[typing.Any]:
        return self._data.__iter__()

    def __len__(self) -> int:
        return self._data.__len__()

    def __ne__(self, other: typing.Any) -> bool:
        return not self.__eq__(other)

    def keys(self) -> typing.KeysView[str]:
        return self._data.keys()

    def values(self) -> typing.ValuesView[typing.Any]:
        return self._data.values()

    def items(self) -> typing.ItemsView[str, typing.Any]:
        return self._data.items()

    def get(self, key: str, default: typing.Any = None) -> typing.Any:
        try:
            return self[key]
        except KeyError:
            return default

    @typing.overload
    def pop(self, key: str) -> typing.Any:
        ...

    @typing.overload
    def pop(self, key: str, default: _T) -> _T:
        ...

    @typing.overload
    def pop(self, key: str, default: typing.Any) -> typing.Any:
        ...

    def pop(self, key: str, default: typing.Any = _UNSET) -> typing.Any:
        if default is _UNSET:
            return self._data.pop(key)
        return self._data.pop(key, default)

    def popitem(self) -> typing.Tuple[str, typing.Any]:
        return self._data.popitem()

    def clear(self) -> None:
        self._data.clear()

    def update(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        self._data.update(*args, **kwargs)

    @typing.overload
    def setdefault(self, key: str, default: None = None) -> None:
        ...

    @typing.overload
    def setdefault(self, key: str, default: typing.Any) -> typing.Any:
        ...

    def setdefault(self, key: str, default: typing.Any = _UNSET) -> typing.Any:
        if default is _UNSET:
            return self._data.setdefault(key)
        return self._data.setdefault(key, default)

    def __eq__(self, other: typing.Any) -> bool:
        try:
            other_model = self.__class__(other)
        except Exception:
            return False
        return self._data == other_model._data

    def __repr__(self) -> str:
        return str(self._data)


def _is_model(obj: typing.Any) -> bool:
    return getattr(obj, "_is_model", False)


def _serialize(o, format: typing.Optional[str] = None):  # pylint: disable=too-many-return-statements
    if isinstance(o, list):
        return [_serialize(x, format) for x in o]
    if isinstance(o, dict):
        return {k: _serialize(v, format) for k, v in o.items()}
    if isinstance(o, set):
        return {_serialize(x, format) for x in o}
    if isinstance(o, tuple):
        return tuple(_serialize(x, format) for x in o)
    if isinstance(o, (bytes, bytearray)):
        return _serialize_bytes(o, format)
    if isinstance(o, decimal.Decimal):
        return float(o)
    if isinstance(o, enum.Enum):
        return o.value
    try:
        # First try datetime.datetime
        return _serialize_datetime(o, format)
    except AttributeError:
        pass
    # Last, try datetime.timedelta
    try:
        return _timedelta_as_isostr(o)
    except AttributeError:
        # This will be raised when it hits value.total_seconds in the method above
        pass
    return o


def _get_rest_field(
    attr_to_rest_field: typing.Dict[str, "_RestField"], rest_name: str
) -> typing.Optional["_RestField"]:
    try:
        return next(rf for rf in attr_to_rest_field.values() if rf._rest_name == rest_name)
    except StopIteration:
        return None


def _create_value(rf: typing.Optional["_RestField"], value: typing.Any) -> typing.Any:
    if not rf:
        return _serialize(value, None)
    if rf._is_multipart_file_input:
        return value
    if rf._is_model:
        return _deserialize(rf._type, value)
    return _serialize(value, rf._format)


class Model(_MyMutableMapping):
    _is_model = True

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        class_name = self.__class__.__name__
        if len(args) > 1:
            raise TypeError(f"{class_name}.__init__() takes 2 positional arguments but {len(args) + 1} were given")
        dict_to_pass = {
            rest_field._rest_name: rest_field._default
            for rest_field in self._attr_to_rest_field.values()
            if rest_field._default is not _UNSET
        }
        if args:
            dict_to_pass.update(
                {k: _create_value(_get_rest_field(self._attr_to_rest_field, k), v) for k, v in args[0].items()}
            )
        else:
            non_attr_kwargs = [k for k in kwargs if k not in self._attr_to_rest_field]
            if non_attr_kwargs:
                # actual type errors only throw the first wrong keyword arg they see, so following that.
                raise TypeError(f"{class_name}.__init__() got an unexpected keyword argument '{non_attr_kwargs[0]}'")
            dict_to_pass.update(
                {
                    self._attr_to_rest_field[k]._rest_name: _create_value(self._attr_to_rest_field[k], v)
                    for k, v in kwargs.items()
                    if v is not None
                }
            )
        super().__init__(dict_to_pass)

    def copy(self) -> "Model":
        return Model(self.__dict__)

    def __new__(cls, *args: typing.Any, **kwargs: typing.Any) -> Self:  # pylint: disable=unused-argument
        # we know the last three classes in mro are going to be 'Model', 'dict', and 'object'
        mros = cls.__mro__[:-3][::-1]  # ignore model, dict, and object parents, and reverse the mro order
        attr_to_rest_field: typing.Dict[str, _RestField] = {  # map attribute name to rest_field property
            k: v for mro_class in mros for k, v in mro_class.__dict__.items() if k[0] != "_" and hasattr(v, "_type")
        }
        annotations = {
            k: v
            for mro_class in mros
            if hasattr(mro_class, "__annotations__")  # pylint: disable=no-member
            for k, v in mro_class.__annotations__.items()  # pylint: disable=no-member
        }
        for attr, rf in attr_to_rest_field.items():
            rf._module = cls.__module__
            if not rf._type:
                rf._type = rf._get_deserialize_callable_from_annotation(annotations.get(attr, None))
            if not rf._rest_name_input:
                rf._rest_name_input = attr
        cls._attr_to_rest_field: typing.Dict[str, _RestField] = dict(attr_to_rest_field.items())

        return super().__new__(cls)  # pylint: disable=no-value-for-parameter

    def __init_subclass__(cls, discriminator: typing.Optional[str] = None) -> None:
        for base in cls.__bases__:
            if hasattr(base, "__mapping__"):  # pylint: disable=no-member
                base.__mapping__[discriminator or cls.__name__] = cls  # type: ignore  # pylint: disable=no-member

    @classmethod
    def _get_discriminator(cls, exist_discriminators) -> typing.Optional[str]:
        for v in cls.__dict__.values():
            if (
                isinstance(v, _RestField) and v._is_discriminator and v._rest_name not in exist_discriminators
            ):  # pylint: disable=protected-access
                return v._rest_name  # pylint: disable=protected-access
        return None

    @classmethod
    def _deserialize(cls, data, exist_discriminators):
        if not hasattr(cls, "__mapping__"):  # pylint: disable=no-member
            return cls(data)
        discriminator = cls._get_discriminator(exist_discriminators)
        exist_discriminators.append(discriminator)
        mapped_cls = cls.__mapping__.get(data.get(discriminator), cls)  # pyright: ignore # pylint: disable=no-member
        if mapped_cls == cls:
            return cls(data)
        return mapped_cls._deserialize(data, exist_discriminators)  # pylint: disable=protected-access

    def as_dict(self, *, exclude_readonly: bool = False) -> typing.Dict[str, typing.Any]:
        """Return a dict that can be JSONify using json.dump.

        :keyword bool exclude_readonly: Whether to remove the readonly properties.
        :returns: A dict JSON compatible object
        :rtype: dict
        """

        result = {}
        if exclude_readonly:
            readonly_props = [p._rest_name for p in self._attr_to_rest_field.values() if _is_readonly(p)]
        for k, v in self.items():
            if exclude_readonly and k in readonly_props:  # pyright: ignore
                continue
            is_multipart_file_input = False
            try:
                is_multipart_file_input = next(
                    rf for rf in self._attr_to_rest_field.values() if rf._rest_name == k
                )._is_multipart_file_input
            except StopIteration:
                pass
            result[k] = v if is_multipart_file_input else Model._as_dict_value(v, exclude_readonly=exclude_readonly)
        return result

    @staticmethod
    def _as_dict_value(v: typing.Any, exclude_readonly: bool = False) -> typing.Any:
        if v is None or isinstance(v, _Null):
            return None
        if isinstance(v, (list, tuple, set)):
            return type(v)(Model._as_dict_value(x, exclude_readonly=exclude_readonly) for x in v)
        if isinstance(v, dict):
            return {dk: Model._as_dict_value(dv, exclude_readonly=exclude_readonly) for dk, dv in v.items()}
        return v.as_dict(exclude_readonly=exclude_readonly) if hasattr(v, "as_dict") else v


def _get_deserialize_callable_from_annotation(  # pylint: disable=R0911, R0915, R0912
    annotation: typing.Any,
    module: typing.Optional[str],
    rf: typing.Optional["_RestField"] = None,
) -> typing.Optional[typing.Callable[[typing.Any], typing.Any]]:
    if not annotation or annotation in [int, float]:
        return None

    # is it a type alias?
    if isinstance(annotation, str):
        if module is not None:
            annotation = _get_type_alias_type(module, annotation)

    # is it a forward ref / in quotes?
    if isinstance(annotation, (str, typing.ForwardRef)):
        try:
            model_name = annotation.__forward_arg__  # type: ignore
        except AttributeError:
            model_name = annotation
        if module is not None:
            annotation = _get_model(module, model_name)

    try:
        if module and _is_model(annotation):
            if rf:
                rf._is_model = True

            def _deserialize_model(model_deserializer: typing.Optional[typing.Callable], obj):
                if _is_model(obj):
                    return obj
                return _deserialize(model_deserializer, obj)

            return functools.partial(_deserialize_model, annotation)  # pyright: ignore
    except Exception:
        pass

    # is it a literal?
    try:
        if annotation.__origin__ is typing.Literal:  # pyright: ignore
            return None
    except AttributeError:
        pass

    # is it optional?
    try:
        if any(a for a in annotation.__args__ if a == type(None)):  # pyright: ignore
            if_obj_deserializer = _get_deserialize_callable_from_annotation(
                next(a for a in annotation.__args__ if a != type(None)), module, rf  # pyright: ignore
            )

            def _deserialize_with_optional(if_obj_deserializer: typing.Optional[typing.Callable], obj):
                if obj is None:
                    return obj
                return _deserialize_with_callable(if_obj_deserializer, obj)

            return functools.partial(_deserialize_with_optional, if_obj_deserializer)
    except AttributeError:
        pass

    if getattr(annotation, "__origin__", None) is typing.Union:
        # initial ordering is we make `string` the last deserialization option, because it is often them most generic
        deserializers = [
            _get_deserialize_callable_from_annotation(arg, module, rf)
            for arg in sorted(
                annotation.__args__, key=lambda x: hasattr(x, "__name__") and x.__name__ == "str"  # pyright: ignore
            )
        ]

        def _deserialize_with_union(deserializers, obj):
            for deserializer in deserializers:
                try:
                    return _deserialize(deserializer, obj)
                except DeserializationError:
                    pass
            raise DeserializationError()

        return functools.partial(_deserialize_with_union, deserializers)

    try:
        if annotation._name == "Dict":  # pyright: ignore
            value_deserializer = _get_deserialize_callable_from_annotation(
                annotation.__args__[1], module, rf  # pyright: ignore
            )

            def _deserialize_dict(
                value_deserializer: typing.Optional[typing.Callable],
                obj: typing.Dict[typing.Any, typing.Any],
            ):
                if obj is None:
                    return obj
                return {k: _deserialize(value_deserializer, v, module) for k, v in obj.items()}

            return functools.partial(
                _deserialize_dict,
                value_deserializer,
            )
    except (AttributeError, IndexError):
        pass
    try:
        if annotation._name in ["List", "Set", "Tuple", "Sequence"]:  # pyright: ignore
            if len(annotation.__args__) > 1:  # pyright: ignore

                def _deserialize_multiple_sequence(
                    entry_deserializers: typing.List[typing.Optional[typing.Callable]],
                    obj,
                ):
                    if obj is None:
                        return obj
                    return type(obj)(
                        _deserialize(deserializer, entry, module)
                        for entry, deserializer in zip(obj, entry_deserializers)
                    )

                entry_deserializers = [
                    _get_deserialize_callable_from_annotation(dt, module, rf)
                    for dt in annotation.__args__  # pyright: ignore
                ]
                return functools.partial(_deserialize_multiple_sequence, entry_deserializers)
            deserializer = _get_deserialize_callable_from_annotation(
                annotation.__args__[0], module, rf  # pyright: ignore
            )

            def _deserialize_sequence(
                deserializer: typing.Optional[typing.Callable],
                obj,
            ):
                if obj is None:
                    return obj
                return type(obj)(_deserialize(deserializer, entry, module) for entry in obj)

            return functools.partial(_deserialize_sequence, deserializer)
    except (TypeError, IndexError, AttributeError, SyntaxError):
        pass

    def _deserialize_default(
        deserializer,
        obj,
    ):
        if obj is None:
            return obj
        try:
            return _deserialize_with_callable(deserializer, obj)
        except Exception:
            pass
        return obj

    if get_deserializer(annotation, rf):
        return functools.partial(_deserialize_default, get_deserializer(annotation, rf))

    return functools.partial(_deserialize_default, annotation)


def _deserialize_with_callable(
    deserializer: typing.Optional[typing.Callable[[typing.Any], typing.Any]],
    value: typing.Any,
):
    try:
        if value is None or isinstance(value, _Null):
            return None
        if deserializer is None:
            return value
        if isinstance(deserializer, CaseInsensitiveEnumMeta):
            try:
                return deserializer(value)
            except ValueError:
                # for unknown value, return raw value
                return value
        if isinstance(deserializer, type) and issubclass(deserializer, Model):
            return deserializer._deserialize(value, [])
        return typing.cast(typing.Callable[[typing.Any], typing.Any], deserializer)(value)
    except Exception as e:
        raise DeserializationError() from e


def _deserialize(
    deserializer: typing.Any,
    value: typing.Any,
    module: typing.Optional[str] = None,
    rf: typing.Optional["_RestField"] = None,
    format: typing.Optional[str] = None,
) -> typing.Any:
    if isinstance(value, PipelineResponse):
        value = value.http_response.json()
    if rf is None and format:
        rf = _RestField(format=format)
    if not isinstance(deserializer, functools.partial):
        deserializer = _get_deserialize_callable_from_annotation(deserializer, module, rf)
    return _deserialize_with_callable(deserializer, value)


class _RestField:
    def __init__(
        self,
        *,
        name: typing.Optional[str] = None,
        type: typing.Optional[typing.Callable] = None,  # pylint: disable=redefined-builtin
        is_discriminator: bool = False,
        visibility: typing.Optional[typing.List[str]] = None,
        default: typing.Any = _UNSET,
        format: typing.Optional[str] = None,
        is_multipart_file_input: bool = False,
    ):
        self._type = type
        self._rest_name_input = name
        self._module: typing.Optional[str] = None
        self._is_discriminator = is_discriminator
        self._visibility = visibility
        self._is_model = False
        self._default = default
        self._format = format
        self._is_multipart_file_input = is_multipart_file_input

    @property
    def _class_type(self) -> typing.Any:
        return getattr(self._type, "args", [None])[0]

    @property
    def _rest_name(self) -> str:
        if self._rest_name_input is None:
            raise ValueError("Rest name was never set")
        return self._rest_name_input

    def __get__(self, obj: Model, type=None):  # pylint: disable=redefined-builtin
        # by this point, type and rest_name will have a value bc we default
        # them in __new__ of the Model class
        item = obj.get(self._rest_name)
        if item is None:
            return item
        if self._is_model:
            return item
        return _deserialize(self._type, _serialize(item, self._format), rf=self)

    def __set__(self, obj: Model, value) -> None:
        if value is None:
            # we want to wipe out entries if users set attr to None
            try:
                obj.__delitem__(self._rest_name)
            except KeyError:
                pass
            return
        if self._is_model:
            if not _is_model(value):
                value = _deserialize(self._type, value)
            obj.__setitem__(self._rest_name, value)
            return
        obj.__setitem__(self._rest_name, _serialize(value, self._format))

    def _get_deserialize_callable_from_annotation(
        self, annotation: typing.Any
    ) -> typing.Optional[typing.Callable[[typing.Any], typing.Any]]:
        return _get_deserialize_callable_from_annotation(annotation, self._module, self)


def rest_field(
    *,
    name: typing.Optional[str] = None,
    type: typing.Optional[typing.Callable] = None,  # pylint: disable=redefined-builtin
    visibility: typing.Optional[typing.List[str]] = None,
    default: typing.Any = _UNSET,
    format: typing.Optional[str] = None,
    is_multipart_file_input: bool = False,
) -> typing.Any:
    return _RestField(
        name=name,
        type=type,
        visibility=visibility,
        default=default,
        format=format,
        is_multipart_file_input=is_multipart_file_input,
    )


def rest_discriminator(
    *,
    name: typing.Optional[str] = None,
    type: typing.Optional[typing.Callable] = None,  # pylint: disable=redefined-builtin
) -> typing.Any:
    return _RestField(name=name, type=type, is_discriminator=True)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access
# enable protected access for protected helper functions

import copy
from collections import OrderedDict
from enum import Enum as PyEnum
from enum import EnumMeta
from inspect import Parameter, getmro, signature
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast

from typing_extensions import Annotated, Literal, TypeAlias

from azure.ai.ml.constants._component import IOConstants
from azure.ai.ml.exceptions import UserErrorException

# avoid circular import error
if TYPE_CHECKING:
    from .input import Input
    from .output import Output

SUPPORTED_RETURN_TYPES_PRIMITIVE = list(IOConstants.PRIMITIVE_TYPE_2_STR.keys())

Annotation: TypeAlias = Union[str, Type, Annotated[Any, Any], None]  # type: ignore


def is_group(obj: object) -> bool:
    """Return True if obj is a group or an instance of a parameter group class.

    :param obj: The object to check.
    :type obj: Any
    :return: True if obj is a group or an instance, False otherwise.
    :rtype: bool
    """
    return hasattr(obj, IOConstants.GROUP_ATTR_NAME)


def _get_annotation_by_value(val: Any) -> Union["Input", Type["Input"]]:
    # TODO: we'd better remove this potential recursive import
    from .enum_input import EnumInput
    from .input import Input

    annotation: Any = None

    def _is_dataset(data: Any) -> bool:
        from azure.ai.ml.entities._job.job_io_mixin import JobIOMixin

        DATASET_TYPES = JobIOMixin
        return isinstance(data, DATASET_TYPES)

    if _is_dataset(val):
        annotation = Input
    elif val is Parameter.empty or val is None:
        # If no default value or default is None, create val as the basic parameter type,
        # it could be replaced using component parameter definition.
        annotation = Input._get_default_unknown_input()
    elif isinstance(val, PyEnum):
        # Handle enum values
        annotation = EnumInput(enum=val.__class__)
    else:
        _new_annotation = _get_annotation_cls_by_type(type(val), raise_error=False)
        if not _new_annotation:
            # Fall back to default
            annotation = Input._get_default_unknown_input()
        else:
            return _new_annotation
    return cast(Union["Input", Type["Input"]], annotation)


def _get_annotation_cls_by_type(
    t: type, raise_error: bool = False, optional: Optional[bool] = None
) -> Optional["Input"]:
    # TODO: we'd better remove this potential recursive import
    from .input import Input

    cls = Input._get_input_by_type(t, optional=optional)
    if cls is None and raise_error:
        raise UserErrorException(f"Can't convert type {t} to azure.ai.ml.Input")
    return cls


# pylint: disable=too-many-statements
def _get_param_with_standard_annotation(
    cls_or_func: Union[Callable, Type], is_func: bool = False, skip_params: Optional[List[str]] = None
) -> Dict[str, Union[Annotation, "Input", "Output"]]:
    """Standardize function parameters or class fields with dsl.types annotation.

    :param cls_or_func: Either a class or a function
    :type cls_or_func: Union[Callable, Type]
    :param is_func: Whether `cls_or_func` is a function. Defaults to False.
    :type is_func: bool
    :param skip_params:
    :type skip_params: Optional[List[str]]
    :return: A dictionary of field annotations
    :rtype: Dict[str, Union[Annotation, "Input", "Output"]]
    """
    # TODO: we'd better remove this potential recursive import
    from .group_input import GroupInput
    from .input import Input
    from .output import Output

    def _is_dsl_type_cls(t: Any) -> bool:
        if type(t) is not type:  # pylint: disable=unidiomatic-typecheck
            return False
        return issubclass(t, (Input, Output))

    def _is_dsl_types(o: object) -> bool:
        return _is_dsl_type_cls(type(o))

    def _get_fields(annotations: Dict) -> Dict:
        """Return field names to annotations mapping in class.

        :param annotations: The annotations
        :type annotations: Dict[str, Union[Annotation, Input, Output]]
        :return: The field dict
        :rtype: Dict[str, Union[Annotation, Input, Output]]
        """
        annotation_fields = OrderedDict()
        for name, annotation in annotations.items():
            # Skip return type
            if name == "return":
                continue
            # Handle EnumMeta annotation
            if isinstance(annotation, EnumMeta):
                from .enum_input import EnumInput

                annotation = EnumInput(type="string", enum=annotation)
            # Handle Group annotation
            if is_group(annotation):
                _deep_copy: GroupInput = copy.deepcopy(getattr(annotation, IOConstants.GROUP_ATTR_NAME))
                annotation = _deep_copy
            # Try creating annotation by type when got like 'param: int'
            if not _is_dsl_type_cls(annotation) and not _is_dsl_types(annotation):
                origin_annotation = annotation
                annotation = cast(Input, _get_annotation_cls_by_type(annotation, raise_error=False))
                if not annotation:
                    msg = f"Unsupported annotation type {origin_annotation!r} for parameter {name!r}."
                    raise UserErrorException(msg)
            annotation_fields[name] = annotation
        return annotation_fields

    def _merge_field_keys(
        annotation_fields: Dict[str, Union[Annotation, Input, Output]], defaults_dict: Dict[str, Any]
    ) -> List[str]:
        """Merge field keys from annotations and cls dict to get all fields in class.

        :param annotation_fields: The field annotations
        :type annotation_fields: Dict[str, Union[Annotation, Input, Output]]
        :param defaults_dict: The map of variable name to default value
        :type defaults_dict: Dict[str, Any]
        :return: A list of field keys
        :rtype: List[str]
        """
        anno_keys = list(annotation_fields.keys())
        dict_keys = defaults_dict.keys()
        if not dict_keys:
            return anno_keys
        return [*anno_keys, *[key for key in dict_keys if key not in anno_keys]]

    def _update_annotation_with_default(
        anno: Union[Annotation, Input, Output], name: str, default: Any
    ) -> Union[Annotation, Input, Output]:
        """Create annotation if is type class and update the default.

        :param anno: The annotation
        :type anno: Union[Annotation, Input, Output]
        :param name: The port name
        :type name: str
        :param default: The default value
        :type default: Any
        :return: The updated annotation
        :rtype: Union[Annotation, Input, Output]
        """
        # Create instance if is type class
        complete_annotation = anno
        if _is_dsl_type_cls(anno):
            if anno is not None and not isinstance(anno, (str, Input, Output)):
                complete_annotation = anno()
        if complete_annotation is not None and not isinstance(complete_annotation, str):
            complete_annotation._port_name = name
        if default is Input._EMPTY:
            return complete_annotation
        if isinstance(complete_annotation, Input):
            # Non-parameter Input has no default attribute
            if complete_annotation._is_primitive_type and complete_annotation.default is not None:
                # logger.warning(
                #     f"Warning: Default value of f{complete_annotation.name!r} is set twice: "
                #     f"{complete_annotation.default!r} and {default!r}, will use {default!r}"
                # )
                pass
            complete_annotation._update_default(default)
        if isinstance(complete_annotation, Output) and default is not None:
            msg = (
                f"Default value of Output {complete_annotation._port_name!r} cannot be set:"
                f"Output has no default value."
            )
            raise UserErrorException(msg)
        return complete_annotation

    def _update_fields_with_default(
        annotation_fields: Dict[str, Union[Annotation, Input, Output]], defaults_dict: Dict[str, Any]
    ) -> Dict[str, Union[Annotation, Input, Output]]:
        """Use public values in class dict to update annotations.

        :param annotation_fields: The field annotations
        :type annotation_fields: Dict[str, Union[Annotation, Input, Output]]
        :param defaults_dict: A map of variable name to default value
        :type defaults_dict: Dict[str, Any]
        :return: List of field names
        :rtype: List[str]
        """
        all_fields = OrderedDict()
        all_filed_keys = _merge_field_keys(annotation_fields, defaults_dict)
        for name in all_filed_keys:
            # Get or create annotation
            annotation = (
                annotation_fields[name]
                if name in annotation_fields
                else _get_annotation_by_value(defaults_dict.get(name, Input._EMPTY))
            )
            # Create annotation if is class type and update default
            annotation = _update_annotation_with_default(annotation, name, defaults_dict.get(name, Input._EMPTY))
            all_fields[name] = annotation
        return all_fields

    def _merge_and_reorder(
        inherited_fields: Dict[str, Union[Annotation, Input, Output]],
        cls_fields: Dict[str, Union[Annotation, Input, Output]],
    ) -> Dict[str, Union[Annotation, Input, Output]]:
        """Merge inherited fields with cls fields.

        The order inside each part will not be changed. Order will be:

        {inherited_no_default_fields} + {cls_no_default_fields} + {inherited_default_fields} + {cls_default_fields}.


        :param inherited_fields: The inherited fields
        :type inherited_fields: Dict[str, Union[Annotation, Input, Output]]
        :param cls_fields: The class fields
        :type cls_fields: Dict[str, Union[Annotation, Input, Output]]
        :return: The merged fields
        :rtype: Dict[str, Union[Annotation, Input, Output]]

        .. admonition:: Additional Note

           :class: note

           If cls overwrite an inherited no default field with default, it will be put in the
           cls_default_fields part and deleted from inherited_no_default_fields:

           .. code-block:: python

              @dsl.group
              class SubGroup:
                  int_param0: Integer
                  int_param1: int

              @dsl.group
              class Group(SubGroup):
                  int_param3: Integer
                  int_param1: int = 1

           The init function of Group will be `def __init__(self, *, int_param0, int_param3, int_param1=1)`.
        """

        def _split(
            _fields: Dict[str, Union[Annotation, Input, Output]]
        ) -> Tuple[Dict[str, Union[Annotation, Input, Output]], Dict[str, Union[Annotation, Input, Output]]]:
            """Split fields to two parts from the first default field.

            :param _fields: The fields
            :type _fields: Dict[str, Union[Annotation, Input, Output]]
            :return: A 2-tuple of (fields with no defaults, fields with defaults)
            :rtype: Tuple[Dict[str, Union[Annotation, Input, Output]], Dict[str, Union[Annotation, Input, Output]]]
            """
            _no_defaults_fields, _defaults_fields = {}, {}
            seen_default = False
            for key, val in _fields.items():
                if val is not None and not isinstance(val, str):
                    if val.get("default", None) or seen_default:
                        seen_default = True
                        _defaults_fields[key] = val
                    else:
                        _no_defaults_fields[key] = val
            return _no_defaults_fields, _defaults_fields

        inherited_no_default, inherited_default = _split(inherited_fields)
        cls_no_default, cls_default = _split(cls_fields)
        # Cross comparison and delete from inherited_fields if same key appeared in cls_fields
        # pylint: disable=consider-iterating-dictionary
        for key in cls_default.keys():
            if key in inherited_no_default.keys():
                del inherited_no_default[key]
        for key in cls_no_default.keys():
            if key in inherited_default.keys():
                del inherited_default[key]
        return OrderedDict(
            {
                **inherited_no_default,
                **cls_no_default,
                **inherited_default,
                **cls_default,
            }
        )

    def _get_inherited_fields() -> Dict[str, Union[Annotation, Input, Output]]:
        """Get all fields inherited from @group decorated base classes.

        :return: The field dict
        :rtype: Dict[str, Union[Annotation, Input, Output]]
        """
        # Return value of _get_param_with_standard_annotation
        _fields: Dict[str, Union[Annotation, Input, Output]] = OrderedDict({})
        if is_func:
            return _fields
        # In reversed order so that more derived classes
        # override earlier field definitions in base classes.
        if isinstance(cls_or_func, type):
            for base in cls_or_func.__mro__[-1:0:-1]:
                if is_group(base):
                    # merge and reorder fields from current base with previous
                    _fields = _merge_and_reorder(
                        _fields, copy.deepcopy(getattr(base, IOConstants.GROUP_ATTR_NAME).values)
                    )
        return _fields

    skip_params = skip_params or []
    inherited_fields = _get_inherited_fields()
    # From annotations get field with type
    annotations: Dict[str, Annotation] = getattr(cls_or_func, "__annotations__", {})
    annotations = {k: v for k, v in annotations.items() if k not in skip_params}
    annotations = _update_io_from_mldesigner(annotations)
    annotation_fields = _get_fields(annotations)
    defaults_dict: Dict[str, Any] = {}
    # Update fields use class field with defaults from class dict or signature(func).paramters
    if not is_func:
        # Only consider public fields in class dict
        defaults_dict = {
            key: val for key, val in cls_or_func.__dict__.items() if not key.startswith("_") and key not in skip_params
        }
    else:
        # Infer parameter type from value if is function
        defaults_dict = {
            key: val.default
            for key, val in signature(cls_or_func).parameters.items()
            if key not in skip_params and val.kind != val.VAR_KEYWORD
        }
    fields = _update_fields_with_default(annotation_fields, defaults_dict)
    all_fields = _merge_and_reorder(inherited_fields, fields)
    return all_fields


def _update_io_from_mldesigner(annotations: Dict[str, Annotation]) -> Dict[str, Union[Annotation, "Input", "Output"]]:
    """Translates IOBase from mldesigner package to azure.ml.entities.Input/Output.

    This function depends on:

    * `mldesigner._input_output._IOBase._to_io_entity_args_dict` to translate Input/Output instance annotations
      to IO entities.
    * class names of `mldesigner._input_output` to translate Input/Output class annotations
      to IO entities.

    :param annotations: A map of variable names to annotations
    :type annotations: Dict[str, Annotation]
    :return: Dict with mldesigner IO types converted to azure-ai-ml Input/Output
    :rtype: Dict[str, Union[Annotation, Input, Output]]
    """
    from typing_extensions import get_args, get_origin

    from azure.ai.ml import Input, Output

    from .enum_input import EnumInput

    mldesigner_pkg = "mldesigner"
    param_name = "_Param"
    return_annotation_key = "return"

    def _is_primitive_type(io: type) -> bool:
        """Checks whether type is a primitive type

        :param io: A type
        :type io: type
        :return: Return true if type is subclass of mldesigner._input_output._Param
        :rtype: bool
        """
        return any(io.__module__.startswith(mldesigner_pkg) and item.__name__ == param_name for item in getmro(io))

    def _is_input_or_output_type(io: type, type_str: Literal["Input", "Output", "Meta"]) -> bool:
        """Checks whether a type is an Input or Output type

        :param io: A type
        :type io: type
        :param type_str: The kind of type to check for
        :type type_str: Literal["Input", "Output", "Meta"]
        :return: Return true if type name contains type_str
        :rtype: bool
        """
        if isinstance(io, type) and io.__module__.startswith(mldesigner_pkg):
            if type_str in io.__name__:
                return True
        return False

    result = {}
    for key, io in annotations.items():  # pylint: disable=too-many-nested-blocks
        if isinstance(io, type):
            if _is_input_or_output_type(io, "Input"):
                # mldesigner.Input -> entities.Input
                io = Input
            elif _is_input_or_output_type(io, "Output"):
                # mldesigner.Output -> entities.Output
                io = Output
            elif _is_primitive_type(io):
                io = (
                    Output(type=io.TYPE_NAME)  # type: ignore
                    if key == return_annotation_key
                    else Input(type=io.TYPE_NAME)  # type: ignore
                )
        elif hasattr(io, "_to_io_entity_args_dict"):
            try:
                if _is_input_or_output_type(type(io), "Input"):
                    # mldesigner.Input() -> entities.Input()
                    if io is not None:
                        io = Input(**io._to_io_entity_args_dict())
                elif _is_input_or_output_type(type(io), "Output"):
                    # mldesigner.Output() -> entities.Output()
                    if io is not None:
                        io = Output(**io._to_io_entity_args_dict())
                elif _is_primitive_type(type(io)):
                    if io is not None and not isinstance(io, str):
                        if io._is_enum():
                            io = EnumInput(**io._to_io_entity_args_dict())
                        else:
                            io = (
                                Output(**io._to_io_entity_args_dict())
                                if key == return_annotation_key
                                else Input(**io._to_io_entity_args_dict())
                            )
            except BaseException as e:
                raise UserErrorException(f"Failed to parse {io} to azure-ai-ml Input/Output: {str(e)}") from e
                # Handle Annotated annotation
        elif get_origin(io) is Annotated:
            hint_type, arg, *hint_args = get_args(io)  # pylint: disable=unused-variable
            if hint_type in SUPPORTED_RETURN_TYPES_PRIMITIVE:
                if not _is_input_or_output_type(type(arg), "Meta"):
                    raise UserErrorException(
                        f"Annotated Metadata class only support "
                        f"mldesigner._input_output.Meta, "
                        f"it is {type(arg)} now."
                    )
                if arg.type is not None and arg.type != hint_type:
                    raise UserErrorException(
                        f"Meta class type {arg.type} should be same as Annotated type: " f"{hint_type}"
                    )
                arg.type = hint_type
                io = (
                    Output(**arg._to_io_entity_args_dict())
                    if key == return_annotation_key
                    else Input(**arg._to_io_entity_args_dict())
                )
        result[key] = io
    return result


def _remove_empty_values(data: Any) -> Any:
    """Recursively removes None values from a dict

    :param data: The value to remove None from
    :type data: T
    :return:
      * `data` if `data` is not a dict
      * `data` with None values recursively filtered out if data is a dict
    :rtype: T
    """
    if not isinstance(data, dict):
        return data
    return {k: _remove_empty_values(v) for k, v in data.items() if v is not None}



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

from marshmallow import fields, post_load

from azure.ai.ml._schema.core.schema_meta import PatchedSchemaMeta


class JobResourcesSchema(metaclass=PatchedSchemaMeta):
    instance_types = fields.List(
        fields.Str(), metadata={"description": "The instance type to make available to this job."}
    )

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities import JobResources

        return JobResources(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from typing import Dict, Optional, Union, overload

from ... import Input, Output
from ..._utils.utils import get_all_enum_values_iter
from ...constants import AssetTypes
from ...constants._common import InputTypes
from ...constants._component import ComponentParameterTypes, IOConstants
from .._schema.input_output import SUPPORTED_INTERNAL_PARAM_TYPES

_INPUT_TYPE_ENUM = "enum"
_INPUT_TYPE_ENUM_CAP = "Enum"
_INPUT_TYPE_FLOAT = "float"
_INPUT_TYPE_FLOAT_CAP = "Float"


class InternalInput(Input):
    """Internal input class for internal components only. Comparing to the public Input class, this class has additional
    primitive input types:

    - String
    - Integer
    - Float, float
    - Boolean
    - Enum, enum (new)
    """

    def __init__(self, *, datastore_mode=None, is_resource=None, **kwargs):
        self.datastore_mode = datastore_mode
        self.is_resource = is_resource
        super().__init__(**kwargs)

    @property
    def _allowed_types(self):
        if self._lower_type == _INPUT_TYPE_ENUM:
            return str
        if self._lower_type == _INPUT_TYPE_FLOAT:
            return float
        return IOConstants.PRIMITIVE_STR_2_TYPE.get(self._lower_type, None)

    @property
    def _lower_type(self) -> Optional[str]:
        if isinstance(self.type, str):
            return self.type.lower()
        if self._multiple_types:
            return None
        return self.type.__name__.lower()

    @property
    def _is_primitive_type(self):
        if self._lower_type in [_INPUT_TYPE_ENUM, _INPUT_TYPE_FLOAT]:
            return True
        if self._lower_type in IOConstants.PRIMITIVE_STR_2_TYPE:
            return True
        return super()._is_primitive_type

    def _simple_parse(self, value, _type=None):
        # simple parse is used to parse min, max & optional only, so we don't need to handle enum
        if _type is None:
            _type = self._lower_type
        if _type == _INPUT_TYPE_FLOAT:
            _type = ComponentParameterTypes.NUMBER
        return super()._simple_parse(value, _type)

    def _to_rest_object(self) -> Dict:
        rest_object = super()._to_rest_object()
        if self._lower_type == _INPUT_TYPE_ENUM:
            rest_object["type"] = _INPUT_TYPE_ENUM_CAP
        if self._lower_type == _INPUT_TYPE_FLOAT:
            rest_object["type"] = _INPUT_TYPE_FLOAT_CAP
        return rest_object

    @classmethod
    def _map_from_rest_type(cls, _type):
        mapping_dict = {
            _INPUT_TYPE_ENUM_CAP: _INPUT_TYPE_ENUM,
            _INPUT_TYPE_FLOAT_CAP: _INPUT_TYPE_FLOAT,
        }
        if isinstance(_type, str) and _type in mapping_dict:
            return mapping_dict[_type]
        return super()._map_from_rest_type(_type)

    @classmethod
    def _from_rest_object(cls, obj: Dict) -> "InternalInput":
        obj["type"] = cls._map_from_rest_type(obj["type"])
        return InternalInput(**obj)

    def _get_python_builtin_type_str(self) -> str:
        if self._lower_type in [_INPUT_TYPE_ENUM, _INPUT_TYPE_FLOAT]:
            return self._lower_type
        if self._is_primitive_type:
            return IOConstants.PRIMITIVE_STR_2_TYPE[self._lower_type].__name__  # type: ignore[index]
            # TODO: Bug 2881900
        return super()._get_python_builtin_type_str()

    @overload
    @classmethod
    def _from_base(cls, _input: None) -> None:  # type: ignore[misc]
        ...

    @overload
    @classmethod
    def _from_base(cls, _input: Union[Input, Dict]) -> "InternalInput": ...

    @classmethod
    def _from_base(cls, _input: Optional[Union[Input, Dict]]) -> Optional["InternalInput"]:
        """Cast from Input or Dict to InternalInput.

        Do not guarantee to create a new object.

        :param _input: The base input
        :type _input: Union[Input, Dict]
        :return:
          * None if _input is None
          * InternalInput
        :rtype: Optional["InternalInput"]
        """
        if _input is None:
            return None
        if isinstance(_input, InternalInput):
            return _input
        if isinstance(_input, Input):
            # do force cast directly as there is no new field added in InternalInput
            # need to change the logic if new field is added
            _input.__class__ = InternalInput
            return _input  # type: ignore[return-value]
        return InternalInput(**_input)


def _map_v1_io_type(output_type: str) -> str:
    """Map v1 IO type to v2.

    :param output_type: The v1 IO type
    :type output_type: str
    :return: The v2 IO type name
    :rtype: str
    """

    # TODO: put it in a common place
    def _map_primitive_type(_type: str) -> str:
        """Convert double and float to number type.

        :param _type: A primitive v1 IO type
        :type _type: str
        :return:
          * InputTypes.NUMBER if _type is "double" or "float"
          * The provided type otherwise
        :rtype: str
        """
        _type = _type.lower()
        if _type in ["double", "float"]:
            return InputTypes.NUMBER
        return _type

    if output_type in list(get_all_enum_values_iter(AssetTypes)):
        return output_type
    if output_type in SUPPORTED_INTERNAL_PARAM_TYPES:
        return _map_primitive_type(output_type)
    if output_type in ["AnyFile"]:
        return AssetTypes.URI_FILE
    # Handle AnyDirectory and the other types.
    return AssetTypes.URI_FOLDER


class InternalOutput(Output):
    def __init__(self, *, datastore_mode=None, is_link_mode=None, **kwargs):
        self.datastore_mode = datastore_mode
        self.is_link_mode = is_link_mode
        super().__init__(**kwargs)

    @classmethod
    def _from_base(cls, _output: Union[Output, Dict]) -> Optional["InternalOutput"]:
        if _output is None:
            return None
        if isinstance(_output, InternalOutput):
            return _output
        if isinstance(_output, Output):
            # do force cast directly as there is no new field added in InternalInput
            # need to change the logic if new field is added
            _output.__class__ = InternalOutput
            return _output  # type: ignore[return-value]
        return InternalOutput(**_output)

    def map_pipeline_output_type(self) -> str:
        """Map output type to pipeline output type.

        :return: The pipeline output type
        :rtype: str
        """
        # TODO: call this for node output
        return _map_v1_io_type(self.type)



import sys
from enum import Enum as PyEnum
from io import StringIO

import pytest
from test_utilities.utils import omit_with_wildcard

from azure.ai.ml import Input, load_component
from azure.ai.ml.constants._component import IOConstants
from azure.ai.ml.dsl import pipeline
from azure.ai.ml.dsl._group_decorator import group
from azure.ai.ml.entities._inputs_outputs import GroupInput, Output, is_group
from azure.ai.ml.entities._job.pipeline._io import PipelineInput, _GroupAttrDict
from azure.ai.ml.exceptions import UserErrorException

from .._util import _DSL_TIMEOUT_SECOND


def remove_extra_character(stdout_IO_value: str) -> str:
    stdout_value = stdout_IO_value.replace(" ", "")
    placeholder = "__DOUBLE_NEWLINE__"
    stdout_value = stdout_value.replace("\n\n", placeholder)

    # Replace all single instances of \n with an empty string
    stdout_value = stdout_value.replace("\n", "")

    # Restore double instances of \n\n from the placeholder
    stdout_value = stdout_value.replace(placeholder, "\n\n")
    return stdout_value


@pytest.mark.usefixtures("enable_pipeline_private_preview_features", "enable_private_preview_schema_features")
@pytest.mark.timeout(_DSL_TIMEOUT_SECOND)
@pytest.mark.unittest
@pytest.mark.pipeline_test
class TestDSLGroup:
    def test_validate_conflict_key(self) -> None:
        def validator(keys, assert_valid=True):
            if assert_valid:
                GroupInput.validate_conflict_keys(keys)
                return
            with pytest.raises(Exception):
                GroupInput.validate_conflict_keys(keys)

        validator(["a", "b", "a.b"], assert_valid=False)
        validator(["a.b", "a.b.c"], assert_valid=False)
        validator(["a", "b", "ab"])
        validator(["aa", "a.c"])
        validator(["a.b", "a.c"])
        validator(["a.bc.de", "bc.de"])
        validator(["a.b", "a.bc", "a.c", "aa", "abc", "ac", "a.ac.b", "ab.c"])

    def test_restore_flattened_inputs(self) -> None:
        keys = ["a.b", "a.bc", "a.c", "aa", "abc", "ac", "a.ac.b", "ab.c"]
        definitions = {k: Input(type="string") for k in keys}
        result = GroupInput.restore_flattened_inputs(definitions)
        assert result.keys() == set({"aa", "abc", "ac", "a", "ab"})
        assert isinstance(result["a"], GroupInput)
        assert result["a"].values.keys() == set({"b", "bc", "ac", "c"})
        assert isinstance(result["a"].values["ac"], GroupInput)
        assert result["a"].values["ac"].values.keys() == set({"b"})
        assert isinstance(result["ab"], GroupInput)
        assert result["ab"].values.keys() == set({"c"})
        inputs = {k: PipelineInput(name=k, meta=None) for k in keys}
        result = _GroupAttrDict(GroupInput.restore_flattened_inputs(inputs))
        assert isinstance(result.a, _GroupAttrDict)
        assert isinstance(result.a.ac, _GroupAttrDict)
        assert isinstance(result.a.b, PipelineInput)
        assert isinstance(result.a.bc, PipelineInput)
        assert isinstance(result.a.c, PipelineInput)
        assert isinstance(result.ab, _GroupAttrDict)
        assert isinstance(result.ab.c, PipelineInput)

    def test_auto_generated_functions(self) -> None:
        class EnumOps(PyEnum):
            Option1 = "Option1"
            Option2 = "Option2"

        @group
        class MixedGroup:
            int_param: int
            str_param: str
            enum_param: EnumOps
            str_default_param: Input(type="string") = "test"
            optional_int_param: Input(type="integer", optional=True) = 5

        # __init__ func test
        assert hasattr(MixedGroup, "__init__") is True
        original_out = sys.stdout
        sys.stdout = stdout_str_IO = StringIO()
        help(MixedGroup.__init__)
        assert (
            "__init__(self,*,int_param:int=None,str_param:str=None,enum_param:str=None,"
            "str_default_param:str='test',optional_int_param:int=5)->None"
            in remove_extra_character(stdout_str_IO.getvalue())
        )
        sys.stdout = original_out

        # __repr__ func test
        var = MixedGroup(
            int_param=1, str_param="test-str", enum_param=EnumOps.Option1, str_default_param="op2", optional_int_param=4
        )
        assert (
            "MixedGroup(int_param=1,str_param='test-str',enum_param=<EnumOps.Option1:'Option1'>,"
            "str_default_param='op2',optional_int_param=4)".replace(" ", "") in var.__repr__().replace(" ", "")
        )

        # __set_attribute__ func test
        var.int_param = 2
        assert var.int_param == 2

    def test_group_items(self):
        @group
        class MixedGroup:
            int_param0: Input(type="integer")
            int_param1: int
            int_param2: Input(type="integer", min=1, max=5)
            int_param3 = 3
            int_param4: Input(type="integer", optional=True) = None

        values = list(getattr(MixedGroup, IOConstants.GROUP_ATTR_NAME).values.values())
        assert len(values) == 5
        values = sorted(values, key=lambda item: item["_port_name"])
        for idx, val in enumerate(values):
            assert val.type == "integer"
            if idx < 4:
                assert val.optional is not True
        assert values[2].min == 1
        assert values[2].max == 5
        assert values[4].optional is True
        assert values[1].default is None

        val = MixedGroup(int_param0=0, int_param1=1, int_param2=2)
        assert val.__dict__ == {"int_param0": 0, "int_param1": 1, "int_param2": 2, "int_param4": None, "int_param3": 3}
        val = MixedGroup(int_param0=0, int_param1=1, int_param2=2, int_param3=9, int_param4=9)
        assert val.__dict__ == {"int_param0": 0, "int_param1": 1, "int_param2": 2, "int_param4": 9, "int_param3": 9}

    def test_create_nested_group(self):
        @group
        class SubGroup:
            param: int = 1

        with pytest.raises(ValueError) as e:

            @group
            class ItemGroup:
                group_param: SubGroup = "str"

            ItemGroup()
        assert "Default value must be instance of parameter group" in str(e)

        @group
        class ItemGroup:
            group_param: SubGroup = SubGroup(param=2)

        assert hasattr(ItemGroup, IOConstants.GROUP_ATTR_NAME) is True
        assert isinstance(getattr(ItemGroup, IOConstants.GROUP_ATTR_NAME).values["group_param"], GroupInput)
        assert is_group(ItemGroup.group_param) is True
        var = ItemGroup()
        assert isinstance(var.group_param.param, PipelineInput)
        assert var.group_param.param._data == 2

    def test_group_supported_types(self):
        # Input Output are supported now
        @group
        class Group:
            param: Input(type="uri_folder")

        with pytest.raises(UserErrorException) as e:

            @pipeline
            def my_pipeline(param: Group):
                pass

        assert "Only primitive types can be used as input of group, got uri_folder" in str(e.value)

        @group
        class Group:
            param: Output

        with pytest.raises(UserErrorException) as e:

            @pipeline
            def my_pipeline(param: Group):
                pass

        assert "Output annotation cannot be used in @pipeline." in str(e.value)

        class CustomizedObj:
            pass

        with pytest.raises(UserErrorException) as e:

            @group
            class Group:
                param: CustomizedObj

        assert "Unsupported annotation type" in str(e.value)

    def test_group_inherit(self):
        @group
        class SubGroup:
            int_param0: int
            int_param1: int = 1

        @group
        class Group(SubGroup):
            int_param3: int
            int_param1: int

        test_group = Group(int_param0=0, int_param1=1, int_param3=3)
        assert test_group.int_param0 == 0
        assert test_group.int_param1 == 1
        assert test_group.int_param3 == 3

    def test_group_inherit_parameter_order(self):
        @group
        class SubGroup0:
            int_param0: int
            int_param1: int
            int_param2: int
            int_param3: int = 3

        @group
        class SubGroup1:
            int_param3: int
            int_param4: int
            int_param1: int = 1
            int_param6: int = 6

        # Merge 0 with 1, order should be:
        # [int_param0, int_param2, int_param3, int_param4], [int_param1=1, int_param6=6]

        @group
        class Group(SubGroup1, SubGroup0):
            int_param1: int
            int_param5: int
            int_param2: int = 2
            int_param7: int = 7

        # Merge Group with (merge 0 with 1), order should be:
        # [int_param0, int_param3, int_param4, int_param1, int_param5], [int_param6=6, int_param2=2, int_param7=7]

        params = getattr(Group, IOConstants.GROUP_ATTR_NAME).values
        assert len(params) == 8
        expected_keys = [
            "int_param0",
            "int_param3",
            "int_param4",
            "int_param1",
            "int_param5",
            "int_param6",
            "int_param2",
            "int_param7",
        ]
        for _id, (expected_key, value) in enumerate(zip(expected_keys, list(params.values()))):
            assert value._port_name == expected_key
            if _id > 4:
                assert value.default is not None
                assert value._port_name == f"int_param{value.default}"

    def test_pipeline_with_group(self):
        class EnumOps(PyEnum):
            Option1 = "option1"
            Option2 = "option2"
            Option3 = "option3"

        @group
        class SubParamClass:
            int_param: Input(min=1.0, max=5.0, type="integer")

        @group
        class ParamClass:
            enum_param: EnumOps
            sub: SubParamClass = SubParamClass(int_param=1)

        default_param = ParamClass(enum_param=EnumOps.Option1)

        hello_world_component_yaml = "./tests/test_configs/components/helloworld_component.yml"
        hello_world_component_func = load_component(source=hello_world_component_yaml)

        @pipeline()
        def pipeline_with_group(group: ParamClass, data: int):
            hello_world_component_func(component_in_path=group.enum_param, component_in_number=group.sub.int_param)

        # Create with parameter group
        pipeline_job = pipeline_with_group(default_param, 1)
        job_dict = pipeline_job._to_dict()
        assert job_dict["inputs"] == {"data": 1, "group.enum_param": "option1", "group.sub.int_param": 1}
        assert job_dict["jobs"]["microsoftsamples_command_component_basic"]["inputs"] == {
            "component_in_number": {"path": "${{parent.inputs.group.sub.int_param}}"},
            "component_in_path": {"path": "${{parent.inputs.group.enum_param}}"},
        }

        # Assign new values one by one
        pipeline_job.inputs.group.sub.int_param = 2
        pipeline_job.inputs.group.enum_param = EnumOps.Option2
        job_dict = pipeline_job._to_dict()
        assert job_dict["inputs"] == {"group.sub.int_param": 2, "group.enum_param": "option2", "data": 1}

        # Assign new group values
        assign_param = ParamClass(enum_param=EnumOps.Option3, sub=SubParamClass(int_param=3))
        pipeline_job.inputs.group = assign_param
        job_dict = pipeline_job._to_dict()
        assert job_dict["inputs"] == {"data": 1, "group.enum_param": "option3", "group.sub.int_param": 3}

        # Assign new sub group values
        pipeline_job.inputs.group.sub = SubParamClass(int_param=5)
        job_dict = pipeline_job._to_dict()
        assert job_dict["inputs"] == {"data": 1, "group.enum_param": "option3", "group.sub.int_param": 5}

    def test_pipeline_with_default_group_value(self):
        @group
        class SubParamClass:
            int_param: Input(min=1.0, max=5.0, type="integer") = 1

        @group
        class ParamClass:
            str_param: str = "string_by_default"
            sub: SubParamClass = SubParamClass()

        hello_world_component_yaml = "./tests/test_configs/components/helloworld_component.yml"
        hello_world_component_func = load_component(hello_world_component_yaml)

        @pipeline()
        def pipeline_with_group_default(group: ParamClass = ParamClass()):
            hello_world_component_func(component_in_path=group.str_param, component_in_number=group.sub.int_param)

        pipeline_job = pipeline_with_group_default()
        job_dict = pipeline_job._to_dict()
        assert job_dict["inputs"] == {"group.str_param": "string_by_default", "group.sub.int_param": 1}
        assert job_dict["jobs"]["microsoftsamples_command_component_basic"]["inputs"] == {
            "component_in_number": {"path": "${{parent.inputs.group.sub.int_param}}"},
            "component_in_path": {"path": "${{parent.inputs.group.str_param}}"},
        }

    def test_assign_group_invalid(self):
        @group
        class ParamClass:
            str_param: str = "string_by_default"

        hello_world_component_yaml = "./tests/test_configs/components/helloworld_component.yml"
        hello_world_component_func = load_component(hello_world_component_yaml)

        @pipeline()
        def str_param_pipeline(str_param, group: ParamClass):
            hello_world_component_func(component_in_path=str_param, component_in_number=1)

        with pytest.raises(Exception) as e:
            str_param_pipeline(str_param=ParamClass(), group=None)
        assert "'str_param' is defined as a parameter but got a parameter group as input" in str(e)

        with pytest.raises(Exception) as e:
            str_param_pipeline(str_param="test", group="test")
        assert "'group' is defined as a parameter group but got input 'test' with type '<class 'str'>" in str(e)

        pipeline_job = str_param_pipeline(str_param="test", group=None)
        pipeline_job.inputs.str_param = ParamClass()
        # Note: this is the expected behavior.
        assert (
            pipeline_job._to_dict()["inputs"]["str_param"]
            == "TestDSLGroup.test_assign_group_invalid.<locals>.ParamClass(str_param='string_by_default')"
        )

        pipeline_job.inputs.str_param = "test"
        with pytest.raises(Exception) as e:
            pipeline_job.inputs.group = "test"
        assert "'group' is expected to be a parameter group, but got <class 'str'>." in str(e)

    def test_group_outputs(self):
        # can define group outputs object, but can not use in @pipeline
        @group
        class PortOutputs:
            # TODO(2097468): automatically change it to Input when used in input annotation
            output1: Output(type="uri_file")
            output2: Output(type="uri_folder")

        with pytest.raises(UserErrorException) as e:

            @pipeline
            def my_pipeline(my_inputs: PortOutputs):
                pass

        assert "Output annotation cannot be used in @pipeline." in str(e.value)

        @group
        class PrimitiveOutputs:
            output1: Output(type="string")
            output2: Output(type="integer")
            output3: Output(type="number")
            output4: Output(type="boolean", is_control=True)

        with pytest.raises(UserErrorException) as e:

            @pipeline
            def my_pipeline(my_inputs: PrimitiveOutputs):
                pass

        assert "Output annotation cannot be used in @pipeline." in str(e.value)

    def test_group_port_inputs(self):
        @group
        class PortInputs:
            input1: Input(type="uri_file")
            input2: Input(type="uri_folder")

        with pytest.raises(UserErrorException) as e:
            # TODO(2097478): support port type groups
            @pipeline
            def my_pipeline(my_inputs: PortInputs):
                pass

        assert "Only primitive types can be used as input of group, got uri_file" in str(e.value)

    def test_group_defaults_with_outputs(self):
        @group
        class MixedGroup:
            int_param: int
            str_default_param: Input(type="string") = "test"
            str_param: str
            input_folder: Input(type="uri_folder")
            optional_int_param: Input(type="integer", optional=True) = 5
            output_folder: Output(type="uri_folder")

        assert hasattr(MixedGroup, "__init__") is True
        original_out = sys.stdout
        sys.stdout = stdout_str_IO = StringIO()
        help(MixedGroup.__init__)
        assert (
            "__init__(self,*,"
            "int_param:int=None,"
            "str_default_param:str='test',"
            "str_param:str=None,"
            "input_folder:{'type':'uri_folder'}=None,"
            "optional_int_param:int=5,"
            "output_folder:{'type':'uri_folder'}=None)"
            "->None" in remove_extra_character(stdout_str_IO.getvalue())
        )
        sys.stdout = original_out
        # __repr__ func test
        var = MixedGroup(
            int_param=1, str_param="test-str", input_folder=Input(path="input"), output_folder=Output(path="output")
        )
        assert (
            "MixedGroup("
            "int_param=1,"
            "str_default_param='test',"
            "str_param='test-str',"
            "input_folder={'type':'uri_folder','path':'input'},"
            "optional_int_param=5,"
            "output_folder={'type':'uri_folder','path':'output'})" in var.__repr__().replace(" ", "")
        )

        # __set_attribute__ func test
        var.input_folder = Input(path="new_input")
        assert var.input_folder.path == "new_input"

    def test_group_port_defaults(self):
        # input
        with pytest.raises(UserErrorException) as e:

            @group
            class SubGroup:
                int_param0: Input
                int_param1: Input = Input(path="in1")

        assert "Default value of Input 'int_param1' cannot be set" in str(e.value)

        with pytest.raises(UserErrorException) as e:

            @group
            class SubGroup:
                out_param0: Output
                out_param1: Output = Output(path="out2")

        assert "Default value of Output 'out_param1' cannot be set" in str(e.value)

    @pytest.mark.skip(reason="Input group item .result() is not supported currently.")
    def test_input_group_result(self):
        @group
        class SubParamClass:
            int_param: Input(min=1.0, max=5.0, type="integer") = 1

        @group
        class ParamClass:
            str_param: str
            sub: SubParamClass = SubParamClass()

        hello_world_component_yaml = "./tests/test_configs/components/helloworld_component.yml"
        hello_world_component_func = load_component(hello_world_component_yaml)

        @pipeline()
        def pipeline_with_group(group: ParamClass, int_param: int):
            assert isinstance(group.str_param.result(), str)
            assert isinstance(group.sub.int_param.result(), int)
            assert isinstance(int_param.result(), int)
            node1 = hello_world_component_func(
                component_in_path=group.str_param, component_in_number=group.sub.int_param.result()
            )
            node2 = hello_world_component_func(
                component_in_path=group.str_param.result(), component_in_number=int_param.result()
            )

            return {"output1": node1.outputs.component_out_path, "output2": node2.outputs.component_out_path}

        pipeline_job1 = pipeline_with_group(group=ParamClass(str_param="str_1"), int_param=1)

        common_omit_fields = [
            "jobs.*.componentId",
            "jobs.*._source",
            "jobs.*.properties",
        ]

        rest_pipeline_job = omit_with_wildcard(
            pipeline_job1._to_rest_object().properties.as_dict(), *common_omit_fields
        )

        expected_pipeline_job1 = {}
        assert rest_pipeline_job == expected_pipeline_job1

        pipeline_job2 = pipeline_with_group(group=ParamClass(str_param="str_2"), int_param=1)

        rest_pipeline_job = omit_with_wildcard(
            pipeline_job2._to_rest_object().properties.as_dict(), *common_omit_fields
        )
        expected_pipeline_job2 = {}
        assert rest_pipeline_job == expected_pipeline_job2

    def test_group_outputs_return(self):
        @group
        class PortOutputs:
            output1: Output(type="uri_folder")
            output2: Output(type="uri_folder")

        hello_world_component_yaml = "./tests/test_configs/components/helloworld_component.yml"
        hello_world_component_func = load_component(source=hello_world_component_yaml)

        @pipeline
        def my_pipeline_dict_return() -> PortOutputs:
            node1 = hello_world_component_func(component_in_number=1, component_in_path=Input(path="/a/path/on/ds"))
            node2 = hello_world_component_func(component_in_number=1, component_in_path=Input(path="/a/path/on/ds"))
            return {"output1": node1.outputs.component_out_path, "output2": node2.outputs.component_out_path}

        pipeline_job1 = my_pipeline_dict_return()

        @pipeline
        def my_pipeline() -> PortOutputs:
            node1 = hello_world_component_func(component_in_number=1, component_in_path=Input(path="/a/path/on/ds"))
            node2 = hello_world_component_func(component_in_number=1, component_in_path=Input(path="/a/path/on/ds"))
            return PortOutputs(output1=node1.outputs.component_out_path, output2=node2.outputs.component_out_path)

        pipeline_job2 = my_pipeline()

        assert pipeline_job1._to_dict()["outputs"] == pipeline_job2._to_dict()["outputs"]

    def test_group_outputs_validation(self):
        # test raise validation error if mismatch
        @group
        class PortOutputs:
            output1: Output(type="uri_folder")
            output2: Output(type="uri_folder")

        hello_world_component_yaml = "./tests/test_configs/components/helloworld_component.yml"
        hello_world_component_func = load_component(source=hello_world_component_yaml)

        @pipeline
        def pipeline_count_mismatch() -> PortOutputs:
            node1 = hello_world_component_func(component_in_number=1, component_in_path=Input(path="/a/path/on/ds"))
            return PortOutputs(output1=node1.outputs.component_out_path)

        with pytest.raises(UserErrorException) as e:
            pipeline_count_mismatch()
        assert "Unmatched outputs between actual pipeline output and output in annotation" in str(e.value)

        @group
        class SinglePortOutput:
            output1: Output(type="mltable")

        @pipeline
        def pipeline_type_mismatch() -> SinglePortOutput:
            node1 = hello_world_component_func(component_in_number=1, component_in_path=Input(path="/a/path/on/ds"))
            return SinglePortOutput(output1=node1.outputs.component_out_path)

        with pytest.raises(UserErrorException) as e:
            pipeline_type_mismatch()

        assert "{'type': 'uri_folder'} != annotation output {'type': 'mltable'}" in str(e.value)

        basic_component = load_component(
            source="./tests/test_configs/components/component_with_conditional_output/spec.yaml"
        )

        @group
        class PrimitiveOutputs1:
            output1: Output(type="boolean", is_control=True)
            output2: Output(type="boolean", is_control=True)
            output3: Output(type="boolean")
            output: Output(type="boolean")

        @pipeline
        def pipeline_is_control_mismatch() -> PrimitiveOutputs1:
            node1 = basic_component()
            return PrimitiveOutputs1(
                output1=node1.outputs.output1,
                output2=node1.outputs.output2,
                output3=node1.outputs.output3,
                output=node1.outputs.output,
            )

        pipeline_is_control_mismatch()

    def test_group_outputs_unsupported_annotation(self):
        @group
        class SubOutputs:
            output1: Output(type="uri_folder")
            output2: Output(type="uri_folder")

        @group
        class ParentOutputs:
            output1: SubOutputs

        hello_world_component_yaml = "./tests/test_configs/components/helloworld_component.yml"
        hello_world_component_func = load_component(source=hello_world_component_yaml)

        with pytest.raises(UserErrorException) as e:

            @pipeline
            def my_pipeline() -> ParentOutputs:
                node1 = hello_world_component_func(component_in_number=1)
                return {
                    "output1": node1.outputs.component_out_path,
                }

        assert "Nested group annotation is not supported in pipeline output." in str(e.value)

        with pytest.raises(UserErrorException) as e:

            @group
            class GroupOutputs:
                output1: PipelineInput

        assert "Unsupported annotation type" in str(e.value)

    def test_input_in_output_group(self):
        @group
        class Outputs:
            output1: Input(type="uri_folder")

        hello_world_component_yaml = "./tests/test_configs/components/helloworld_component.yml"
        hello_world_component_func = load_component(source=hello_world_component_yaml)

        @pipeline
        def my_pipeline() -> Outputs:
            node1 = hello_world_component_func(component_in_number=1)
            return Outputs(
                output1=node1.outputs.component_out_path,
            )

        pipeline_job = my_pipeline()
        assert pipeline_job._to_dict()["outputs"] == {"output1": {"type": "uri_folder"}}



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

from typing import Any, Optional


class SourceProcessCodeMetadata(object):
    def __init__(self, *, path: str, process_class: Optional[str] = None, **kwargs: Any):
        self.path = path
        self.process_class = process_class



from typing import Union

import pytest

from azure.ai.ml._restclient.v2024_01_01_preview.models import AutoNCrossValidations, CustomNCrossValidations, JobBase
from azure.ai.ml.automl import classification, forecasting, regression
from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.entities._inputs_outputs import Input
from azure.ai.ml.entities._job.automl.automl_job import AutoMLJob
from azure.ai.ml.entities._job.automl.tabular.classification_job import ClassificationJob
from azure.ai.ml.entities._job.automl.tabular.forecasting_job import ForecastingJob
from azure.ai.ml.entities._job.automl.tabular.regression_job import RegressionJob


@pytest.mark.automl_test
@pytest.mark.unittest
class TestNCrossValidationSettings:
    JOBS = [classification, regression, forecasting]

    def get_job_rest_object(
        self,
        n_cross_validation: Union[str, int, None],
        automl_job: Union[classification, regression, forecasting],
        validation_data_size: float = None,
        from_rest_obj: bool = False,
    ) -> Union[JobBase, ClassificationJob, RegressionJob, ForecastingJob]:
        # Create an AutoML Job

        job = automl_job(
            training_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/train.csv"),
            target_column_name="target",
            enable_model_explainability=True,
            n_cross_validations=n_cross_validation,
            validation_data_size=validation_data_size,
        )

        job.limits = {"timeout_minutes": 60, "max_trials": 100, "max_concurrent_trials": 4}

        if automl_job == forecasting:
            job.set_forecast_settings(
                time_column_name="time",
            )

        rest_obj = job._to_rest_object()

        if from_rest_obj:  # from rest obj
            rest_obj = AutoMLJob._from_rest_object(rest_obj)

        return rest_obj

    def test_auto_n_cv_to_rest(self):
        # Test with auto n cross validations (to_rest_object)
        rest_objs = [
            self.get_job_rest_object("auto", self.JOBS[0]),
            self.get_job_rest_object("auto", self.JOBS[1]),
            self.get_job_rest_object("auto", self.JOBS[2]),
        ]

        for obj in rest_objs:
            assert isinstance(
                obj.properties.task_details.n_cross_validations, AutoNCrossValidations
            ), "N cross validations not an object of AutoNCrossValidations in {} job".format(
                obj.properties.task_details.task_type
            )

    def test_auto_n_cv_from_rest(self):
        # Test with auto n cross validations (from_rest_object)
        rest_objs = [
            self.get_job_rest_object("auto", self.JOBS[0], from_rest_obj=True),
            self.get_job_rest_object("auto", self.JOBS[1], from_rest_obj=True),
            self.get_job_rest_object("auto", self.JOBS[2], from_rest_obj=True),
        ]

        for obj in rest_objs:
            assert obj.n_cross_validations == "auto", "N cross validations incorrectly deserializing in {} job".format(
                obj.task_type
            )

    def test_value_n_cv_to_rest(self):
        # Test with int value n cross validations (to_rest_obj)
        rest_objs = [
            self.get_job_rest_object(2, self.JOBS[0]),
            self.get_job_rest_object(2, self.JOBS[1]),
            self.get_job_rest_object(2, self.JOBS[2]),
        ]

        for obj in rest_objs:
            assert isinstance(
                obj.properties.task_details.n_cross_validations, CustomNCrossValidations
            ), "N cross validations not an object of CustomNCrossValidations in {} job".format(
                obj.properties.task_details.task_type
            )

    def test_value_n_cv_from_rest(self):
        # Test with auto n cross validations (from_rest_object)
        rest_objs = [
            self.get_job_rest_object(2, self.JOBS[0], from_rest_obj=True),
            self.get_job_rest_object(2, self.JOBS[1], from_rest_obj=True),
            self.get_job_rest_object(2, self.JOBS[2], from_rest_obj=True),
        ]

        for obj in rest_objs:
            assert obj.n_cross_validations == 2, "N cross validations incorrectly deserializing in {} job".format(
                obj.task_type
            )

    def test_none_n_cv_to_rest(self):
        # Test with None value n cross validations
        rest_objs = [
            self.get_job_rest_object(None, self.JOBS[0]),
            self.get_job_rest_object(None, self.JOBS[1]),
            self.get_job_rest_object(None, self.JOBS[2]),
        ]

        for obj in rest_objs:
            assert obj.properties.task_details.n_cross_validations is None, "N cross validations not None"

    def test_none_n_cv_from_rest(self):
        # Test with auto n cross validations (from_rest_object)
        rest_objs = [
            self.get_job_rest_object(None, self.JOBS[0], from_rest_obj=True),
            self.get_job_rest_object(None, self.JOBS[1], from_rest_obj=True),
            self.get_job_rest_object(None, self.JOBS[2], from_rest_obj=True),
        ]

        for obj in rest_objs:
            assert obj.n_cross_validations is None, "N cross validations not None in {} job".format(obj.task_type)

    def test_monte_carlo_cv(self):
        # Test with monte carlo cv
        rest_objs = [
            self.get_job_rest_object(10, self.JOBS[0], 0.2),
            self.get_job_rest_object(10, self.JOBS[1], 0.2),
            self.get_job_rest_object(10, self.JOBS[2], 0.2),
        ]

        for obj in rest_objs:
            assert isinstance(
                obj.properties.task_details.n_cross_validations, CustomNCrossValidations
            ), "N cross validations not an object of CustomNCrossValidations in {} job".format(
                obj.properties.task_details.task_type
            )
            assert (
                obj.properties.task_details.validation_data_size == 0.2
            ), "Validation data size not being assigned in {} job".format(obj.properties.task_details.task_type)



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from enum import Enum
from azure.core import CaseInsensitiveEnumMeta


class AliasPathAttributes(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The attributes of the token that the alias path is referring to."""

    NONE = "None"  #: The token that the alias path is referring to has no attributes.
    MODIFIABLE = (
        "Modifiable"  #: The token that the alias path is referring to is modifiable by policies with 'modify' effect.
    )


class AliasPathTokenType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The type of the token that the alias path is referring to."""

    NOT_SPECIFIED = "NotSpecified"  #: The token type is not specified.
    ANY = "Any"  #: The token type can be anything.
    STRING = "String"  #: The token type is string.
    OBJECT = "Object"  #: The token type is object.
    ARRAY = "Array"  #: The token type is array.
    INTEGER = "Integer"  #: The token type is integer.
    NUMBER = "Number"  #: The token type is number.
    BOOLEAN = "Boolean"  #: The token type is boolean.


class AliasPatternType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The type of alias pattern"""

    NOT_SPECIFIED = "NotSpecified"  #: NotSpecified is not allowed.
    EXTRACT = "Extract"  #: Extract is the only allowed value.


class AliasType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The type of the alias."""

    NOT_SPECIFIED = "NotSpecified"  #: Alias type is unknown (same as not providing alias type).
    PLAIN_TEXT = "PlainText"  #: Alias value is not secret.
    MASK = "Mask"  #: Alias value is secret.


class ChangeType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Type of change that will be made to the resource when the deployment is executed."""

    CREATE = "Create"  #: The resource does not exist in the current state but is present in the desired state. The resource will be created when the deployment is executed.
    DELETE = "Delete"  #: The resource exists in the current state and is missing from the desired state. The resource will be deleted when the deployment is executed.
    IGNORE = "Ignore"  #: The resource exists in the current state and is missing from the desired state. The resource will not be deployed or modified when the deployment is executed.
    DEPLOY = "Deploy"  #: The resource exists in the current state and the desired state and will be redeployed when the deployment is executed. The properties of the resource may or may not change.
    NO_CHANGE = "NoChange"  #: The resource exists in the current state and the desired state and will be redeployed when the deployment is executed. The properties of the resource will not change.
    MODIFY = "Modify"  #: The resource exists in the current state and the desired state and will be redeployed when the deployment is executed. The properties of the resource will change.


class DeploymentMode(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The mode that is used to deploy resources. This value can be either Incremental or Complete. In
    Incremental mode, resources are deployed without deleting existing resources that are not
    included in the template. In Complete mode, resources are deployed and existing resources in
    the resource group that are not included in the template are deleted. Be careful when using
    Complete mode as you may unintentionally delete resources.
    """

    INCREMENTAL = "Incremental"
    COMPLETE = "Complete"


class OnErrorDeploymentType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The deployment on error behavior type. Possible values are LastSuccessful and
    SpecificDeployment.
    """

    LAST_SUCCESSFUL = "LastSuccessful"
    SPECIFIC_DEPLOYMENT = "SpecificDeployment"


class PropertyChangeType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The type of property change."""

    CREATE = "Create"  #: The property does not exist in the current state but is present in the desired state. The property will be created when the deployment is executed.
    DELETE = "Delete"  #: The property exists in the current state and is missing from the desired state. It will be deleted when the deployment is executed.
    MODIFY = "Modify"  #: The property exists in both current and desired state and is different. The value of the property will change when the deployment is executed.
    ARRAY = "Array"  #: The property is an array and contains nested changes.


class ProvisioningOperation(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The name of the current provisioning operation."""

    NOT_SPECIFIED = "NotSpecified"  #: The provisioning operation is not specified.
    CREATE = "Create"  #: The provisioning operation is create.
    DELETE = "Delete"  #: The provisioning operation is delete.
    WAITING = "Waiting"  #: The provisioning operation is waiting.
    AZURE_ASYNC_OPERATION_WAITING = (
        "AzureAsyncOperationWaiting"  #: The provisioning operation is waiting Azure async operation.
    )
    RESOURCE_CACHE_WAITING = "ResourceCacheWaiting"  #: The provisioning operation is waiting for resource cache.
    ACTION = "Action"  #: The provisioning operation is action.
    READ = "Read"  #: The provisioning operation is read.
    EVALUATE_DEPLOYMENT_OUTPUT = "EvaluateDeploymentOutput"  #: The provisioning operation is evaluate output.
    DEPLOYMENT_CLEANUP = "DeploymentCleanup"  #: The provisioning operation is cleanup. This operation is part of the 'complete' mode deployment.


class ProvisioningState(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Denotes the state of provisioning."""

    NOT_SPECIFIED = "NotSpecified"
    ACCEPTED = "Accepted"
    RUNNING = "Running"
    READY = "Ready"
    CREATING = "Creating"
    CREATED = "Created"
    DELETING = "Deleting"
    DELETED = "Deleted"
    CANCELED = "Canceled"
    FAILED = "Failed"
    SUCCEEDED = "Succeeded"
    UPDATING = "Updating"


class ResourceIdentityType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The identity type."""

    SYSTEM_ASSIGNED = "SystemAssigned"
    USER_ASSIGNED = "UserAssigned"
    SYSTEM_ASSIGNED_USER_ASSIGNED = "SystemAssigned, UserAssigned"
    NONE = "None"


class TagsPatchOperation(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The operation type for the patch API."""

    REPLACE = "Replace"  #: The 'replace' option replaces the entire set of existing tags with a new set.
    MERGE = "Merge"  #: The 'merge' option allows adding tags with new names and updating the values of tags with existing names.
    DELETE = "Delete"  #: The 'delete' option allows selectively deleting tags based on given names or name/value pairs.


class WhatIfResultFormat(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The format of the What-If results"""

    RESOURCE_ID_ONLY = "ResourceIdOnly"
    FULL_RESOURCE_PAYLOADS = "FullResourcePayloads"



# coding: utf-8

# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: ml_samples_workspace.py
DESCRIPTION:
    These samples demonstrate different ways to configure Workspace and related objects.
USAGE:
    python ml_samples_workspace.py

"""

import os

from ml_samples_compute import handle_resource_exists_error

from azure.ai.ml import MLClient
from azure.identity import DefaultAzureCredential

subscription_id = os.environ["AZURE_SUBSCRIPTION_ID"]
resource_group = os.environ["RESOURCE_GROUP_NAME"]
credential = DefaultAzureCredential()
ml_client = MLClient(credential, subscription_id, resource_group)
ml_client_ws = MLClient(credential, subscription_id, resource_group, workspace_name="test-ws1")


class WorkspaceConfigurationOptions(object):
    def ml_workspace_config_sample_snippets_entities(self):
        # [START load_workspace]
        from azure.ai.ml import load_workspace

        ws = load_workspace(
            "../tests/test_configs/workspace/workspace_min.yaml",
            params_override=[{"description": "loaded from workspace_min.yaml"}],
        )
        # [END load_workspace]

        # [START load_workspace_connection]
        from azure.ai.ml import load_connection

        wps_connection = load_connection(source="../tests/test_configs/connection/snowflake_user_pwd.yaml")
        # [END load_workspace_connection]

        # [START customermanagedkey]
        from azure.ai.ml.entities import CustomerManagedKey, Workspace

        cmk = CustomerManagedKey(
            key_vault="/subscriptions/00000000-1111-2222-3333-444444444444/resourceGroups/test-rg/providers/Microsoft.KeyVault/vaults/vault-name",
            key_uri="https://vault-name.vault.azure.net/keys/key-name/key-version",
        )

        # special bring your own scenario
        byo_cmk = CustomerManagedKey(
            key_vault="/subscriptions/00000000-1111-2222-3333-444444444444/resourceGroups/test-rg/providers/Microsoft.KeyVault/vaults/vault-name",
            key_uri="https://vault-name.vault.azure.net/keys/key-name/key-version",
            cosmosdb_id="/subscriptions/00000000-1111-2222-3333-444444444444/resourceGroups/test-rg/providers/Microsoft.DocumentDB/databaseAccounts/cosmos-name",
            storage_id="/subscriptions/00000000-1111-2222-3333-444444444444/resourceGroups/test-rg/providers/Microsoft.Storage/storageAccounts/storage-name",
            search_id="/subscriptions/00000000-1111-2222-3333-444444444444/resourceGroups/test-rg/providers/Microsoft.Search/searchServices/search-name",
        )

        ws = Workspace(name="ws-name", location="eastus", display_name="My workspace", customer_managed_key=cmk)
        # [END customermanagedkey]

        # [START workspace_managed_network]
        from azure.ai.ml.constants._workspace import FirewallSku
        from azure.ai.ml.entities import (
            FqdnDestination,
            IsolationMode,
            ManagedNetwork,
            PrivateEndpointDestination,
            ServiceTagDestination,
            Workspace,
        )

        # Example private endpoint outbound to a blob
        blobrule = PrivateEndpointDestination(
            name="blobrule",
            service_resource_id="/subscriptions/00000000-1111-2222-3333-444444444444/resourceGroups/test-rg/providers/Microsoft.Storage/storageAccounts/storage-name",
            subresource_target="blob",
            spark_enabled=False,
        )

        # Example service tag rule
        datafactoryrule = ServiceTagDestination(
            name="datafactory", service_tag="DataFactory", protocol="TCP", port_ranges="80, 8080-8089"
        )

        # Example FQDN rule
        pypirule = FqdnDestination(name="pypirule", destination="pypi.org")

        # Example FirewallSku
        # FirewallSku is an optional parameter, when unspecified this will default to FirewallSku.Standard
        firewallSku = FirewallSku.BASIC

        network = ManagedNetwork(
            isolation_mode=IsolationMode.ALLOW_ONLY_APPROVED_OUTBOUND,
            outbound_rules=[blobrule, datafactoryrule, pypirule],
            firewall_sku=firewallSku,
        )

        # Workspace configuration
        ws = Workspace(name="ws-name", location="eastus", managed_network=network)
        # [END workspace_managed_network]

        # [START workspace_managed_network_provision_now]
        from azure.ai.ml.entities import IsolationMode, ManagedNetwork, Workspace

        managed_net = ManagedNetwork(isolation_mode=IsolationMode.ALLOW_INTERNET_OUTBOUND)
        ws = Workspace(name="ws-name", location="eastus", managed_network=managed_net, provision_network_now=True)
        # [END workspace_managed_network_provision_now]

        # [START fqdn_outboundrule]
        from azure.ai.ml.entities import FqdnDestination

        # Example FQDN rule
        pypirule = FqdnDestination(name="rulename", destination="pypi.org")
        # [END fqdn_outboundrule]

        # [START private_endpoint_outboundrule]
        from azure.ai.ml.entities import PrivateEndpointDestination

        # Example private endpoint outbound to a blob
        blobrule = PrivateEndpointDestination(
            name="blobrule",
            service_resource_id="/subscriptions/00000000-1111-2222-3333-444444444444/resourceGroups/test-rg/providers/Microsoft.Storage/storageAccounts/storage-name",
            subresource_target="blob",
            spark_enabled=False,
        )

        # Example private endpoint outbound to an application gateway
        appGwRule = PrivateEndpointDestination(
            name="appGwRule",
            service_resource_id="/subscriptions/00000000-1111-2222-3333-444444444444/resourceGroups/test-rg/providers/Microsoft.Network/applicationGateways/appgw-name",  # cspell:disable-line
            subresource_target="appGwPrivateFrontendIpIPv4",
            spark_enabled=False,
            fqdns=["contoso.com", "contoso2.com"],
        )
        # [END private_endpoint_outboundrule]

        # [START service_tag_outboundrule]
        from azure.ai.ml.entities import ServiceTagDestination

        # Example service tag rule
        datafactoryrule = ServiceTagDestination(
            name="datafactory", service_tag="DataFactory", protocol="TCP", port_ranges="80, 8080-8089"
        )

        # Example service tag rule using custom address prefixes
        customAddressPrefixesRule = ServiceTagDestination(
            name="customAddressPrefixesRule",
            address_prefixes=["168.63.129.16", "10.0.0.0/24"],
            protocol="TCP",
            port_ranges="80, 443, 8080-8089",
        )
        # [END service_tag_outboundrule]

        # [START workspace]
        from azure.ai.ml.entities import Workspace

        ws = Workspace(name="sample-ws", location="eastus", description="a sample workspace object")
        # [END workspace]

        # [START workspace_hub]
        from azure.ai.ml.entities import Hub

        ws = Hub(name="sample-ws", location="eastus", description="a sample workspace hub object")
        # [END workspace_hub]

        # [START workspace_network_access_settings]
        from azure.ai.ml.entities import DefaultActionType, IPRule, NetworkAcls

        # Get existing workspace
        ws = ml_client.workspaces.get("test-ws1")

        # 1. Enabled from all networks
        # Note: default_action should be set to 'Allow', allowing all access.
        ws.public_network_access = "Enabled"
        ws.network_acls = NetworkAcls(default_action=DefaultActionType.ALLOW, ip_rules=[])
        updated_ws = ml_client.workspaces.begin_update(workspace=ws).result()

        # 2. Enabled from selected IP addresses
        # Note: default_action should be set to 'Deny', allowing only specified IPs/ranges
        ws.public_network_access = "Enabled"
        ws.network_acls = NetworkAcls(
            default_action=DefaultActionType.DENY,
            ip_rules=[IPRule(value="103.248.19.87/32"), IPRule(value="103.248.19.86/32")],
        )
        updated_ws = ml_client.workspaces.begin_update(workspace=ws).result()

        # 3. Disabled
        # NetworkAcls IP Rules will reset
        ws.public_network_access = "Disabled"
        updated_ws = ml_client.workspaces.begin_update(workspace=ws).result()
        # [END workspace_network_access_settings]

    @handle_resource_exists_error
    def ml_workspace_config_sample_snippets_operations(self):
        # [START workspace_list]
        from azure.ai.ml.constants import Scope

        # list workspaces in the resource group set in ml_client
        workspaces = ml_client.workspaces.list()
        workspaces = ml_client.workspaces.list(scope=Scope.RESOURCE_GROUP)

        # list workspaces in the subscription set in ml_client
        workspaces = ml_client.workspaces.list(scope=Scope.SUBSCRIPTION)
        # [END workspace_list]

        # [START workspace_get]
        workspace = ml_client.workspaces.get(name="test-ws1")
        # [END workspace_get]

        # [START workspace_get_keys]
        ws_keys = ml_client.workspaces.get_keys(name="test-ws1")
        # [END workspace_get_keys]

        # [START workspace_sync_keys]
        ml_client.workspaces.begin_sync_keys(name="test-ws1")
        # [END workspace_sync_keys]

        # [START workspace_provision_network]
        ml_client.workspaces.begin_provision_network(workspace_name="test-ws1", include_spark=False)
        # [END workspace_provision_network]

        # [START workspace_begin_create]
        from azure.ai.ml.entities import Workspace

        ws = Workspace(
            name="test-ws1",
            description="a test workspace",
            tags={"purpose": "demo"},
            location="eastus",
            resource_group=resource_group,
        )
        ws = ml_client.workspaces.begin_create(workspace=ws).result()
        # [END workspace_begin_create]

        # [START workspace_begin_update]
        ws = ml_client.workspaces.get(name="test-ws1")
        ws.description = "a different description"
        ws = ml_client.workspaces.begin_update(workspace=ws).result()
        # [END workspace_begin_update]

        # [START workspace_begin_delete]
        ml_client.workspaces.begin_delete(name="test-ws", delete_dependent_resources=True, permanently_delete=True)
        # [END workspace_begin_delete]

        # [START workspace_begin_diagnose]
        diagnose_result = ml_client.workspaces.begin_diagnose(name="test-ws1").result()
        # [END workspace_begin_diagnose]

        # [START get_connection]
        from azure.ai.ml import MLClient

        ml_client_ws = MLClient(credential, subscription_id, resource_group, workspace_name="test-ws1")
        connection = ml_client.connections.get(name="test-ws1")
        # [END get_connection]

        # [START create_or_update_connection]
        from azure.ai.ml import MLClient
        from azure.ai.ml.entities import UsernamePasswordConfiguration, WorkspaceConnection

        ml_client_ws = MLClient(credential, subscription_id, resource_group, workspace_name="test-ws")
        wps_connection = WorkspaceConnection(
            name="connection-1",
            type="snowflake",
            target="jdbc:snowflake://<myaccount>.snowflakecomputing.com/?db=<mydb>&warehouse=<mywarehouse>&role=<myrole>",  # cspell:disable-line
            credentials=UsernamePasswordConfiguration(username="XXXXX", password="XXXXXX"),
        )
        connection = ml_client_ws.connections.create_or_update(workspace_connection=wps_connection)
        # [END create_or_update_connection]

        # [START delete_connection]
        from azure.ai.ml import MLClient

        ml_client_ws = MLClient(credential, subscription_id, resource_group, workspace_name="test-ws")
        ml_client_ws.connections.delete(name="connection-1")
        # [END delete_connection]

        # [START list_connection]
        from azure.ai.ml import MLClient

        ml_client_ws = MLClient(credential, subscription_id, resource_group, workspace_name="test-ws")
        ml_client_ws.connections.list(connection_type="git")
        # [END list_connection]

        # [START hub_list]
        from azure.ai.ml.constants import Scope

        # list workspaces in the resource group set in ml_client
        hubs = ml_client.workspace_hubs.list()  # type:ignore
        hubs = ml_client.workspace_hubs.list(scope=Scope.RESOURCE_GROUP)  # type:ignore

        # list workspaces in the subscription set in ml_client
        hubs = ml_client.workspace_hubs.list(scope=Scope.SUBSCRIPTION)  # type:ignore
        # [END hub_list]

        # [START hub_get]
        workspace = ml_client.workspace_hubs.get(name="test-hub1")  # type:ignore
        # [END hub_get]

        # [START hub_begin_create]
        from azure.ai.ml.entities import Hub

        hub = Hub(
            name="test-hub1",
            description="a test hub",
            tags={"purpose": "demo"},
            location="eastus",
            resource_group=resource_group,
        )
        hub = ml_client.workspace_hubs.begin_create(workspace_hub=hub).result()  # type:ignore
        # [END hub_begin_create]

        # [START hub_begin_update]
        hub = ml_client.workspace_hubs.get(name="test-hub1")  # type:ignore
        hub.description = "a different description"
        hub = ml_client.workspace_hubs.begin_update(workspace_hub=hub).result()  # type:ignore
        # [END hub_begin_update]

        # [START hub_begin_delete]
        ml_client.workspace_hubs.begin_delete(  # type:ignore
            name="test-hub1", delete_dependent_resources=True, permanently_delete=True
        )
        # [END hub_begin_delete]

        # [START outbound_rule_list]
        rules = ml_client.workspace_outbound_rules.list(workspace_name="test-ws")
        # [END outbound_rule_list]

        # [START outbound_rule_get]
        rule = ml_client.workspace_outbound_rules.get(workspace_name="test-ws", outbound_rule_name="sample-rule")
        # [END outbound_rule_get]

        # [START outbound_rule_begin_create]
        from azure.ai.ml.entities import FqdnDestination

        fqdn_rule = FqdnDestination(name="rulename", destination="google.com")
        rule = ml_client.workspace_outbound_rules.begin_create(workspace_name="test-ws", rule=fqdn_rule).result()
        # [END outbound_rule_begin_create]

        # [START outbound_rule_begin_update]
        from azure.ai.ml.entities import FqdnDestination

        fqdn_rule = FqdnDestination(name="rulename", destination="linkedin.com")
        rule = ml_client.workspace_outbound_rules.begin_update(workspace_name="test-ws", rule=fqdn_rule).result()
        # [END outbound_rule_begin_update]

        # [START outbound_rule_begin_remove]
        ml_client.workspace_outbound_rules.begin_remove(workspace_name="test-ws", outbound_rule_name="rulename")
        # [END outbound_rule_begin_remove]


if __name__ == "__main__":
    sample = WorkspaceConfigurationOptions()
    sample.ml_workspace_config_sample_snippets_entities()
    sample.ml_workspace_config_sample_snippets_operations()



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from marshmallow import fields

from azure.ai.ml._schema.core.fields import ExperimentalField, NestedField, StringTransformedEnum
from azure.ai.ml._schema.job import BaseJobSchema
from azure.ai.ml._schema.job.input_output_fields_provider import OutputsField
from azure.ai.ml._schema.job_resource_configuration import JobResourceConfigurationSchema
from azure.ai.ml._schema.queue_settings import QueueSettingsSchema
from azure.ai.ml.constants import JobType


class AutoMLJobSchema(BaseJobSchema):
    type = StringTransformedEnum(required=True, allowed_values=JobType.AUTOML)
    environment_id = fields.Str()
    environment_variables = fields.Dict(keys=fields.Str(), values=fields.Str())
    outputs = OutputsField()
    resources = NestedField(JobResourceConfigurationSchema())
    queue_settings = ExperimentalField(NestedField(QueueSettingsSchema))



from pathlib import Path

from azure.ai.ml import Input, command, dsl
from azure.ai.ml.entities import PipelineJob

parent_dir = str(Path(__file__).parent)


def generate_dsl_pipeline() -> PipelineJob:
    environment = "AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33"

    # 1. Construct command job
    train_func = command(
        inputs=dict(
            training_data=Input(path=parent_dir + "/data/"),
            max_epocs=20,
            learning_rate=1.8,
            learning_rate_schedule="time-based",
        ),
        outputs=dict(model_output=None),
        display_name="my-train-job",
        code=parent_dir + "/train_src",
        environment=environment,
        command="python train.py --training_data ${{inputs.training_data}} --max_epocs ${{inputs.max_epocs}} "
        "--learning_rate ${{inputs.learning_rate}} --learning_rate_schedule "
        "${{inputs.learning_rate_schedule}} --model_output ${{outputs.model_output}}",
    )

    score_func = command(
        inputs=dict(
            model_input=Input(path=parent_dir + "/data/"),
            test_data=Input(path=parent_dir + "/data/"),
        ),
        outputs=dict(score_output=None),
        display_name="my-score-job",
        code=parent_dir + "/train_src",
        environment=environment,
        command="python score.py --model_input ${{inputs.model_input}} "
        "--test_data ${{inputs.test_data}} --score_output ${{outputs.score_output}}",
    )

    eval_func = command(
        inputs=dict(
            scoring_result=Input(path=parent_dir + "/data/"),
        ),
        outputs=dict(eval_output=None),
        display_name="my-evaluate-job",
        environment=environment,
        command='echo "hello world"',
    )

    # 2. Construct pipeline
    @dsl.pipeline(
        compute="cpu-cluster",
    )
    def command_job_in_pipeline(
        pipeline_job_training_input,
        pipeline_job_test_input,
        pipeline_job_training_max_epocs,
        pipeline_job_training_learning_rate,
        pipeline_job_learning_rate_schedule,
    ):
        # TODO: the target of this test changed as Input binding is not supported now?
        train_job = train_func(
            training_data=Input(path=parent_dir + "/data/"),
            max_epocs=20,
            learning_rate=1.8,
            learning_rate_schedule="time-based",
        )

        score_job = score_func(
            model_input=train_job.outputs.model_output,
            test_data=pipeline_job_test_input,
        )
        score_job.compute = "gpu-cluster"

        evaluate_job = eval_func(scoring_result=Input(path=parent_dir + "/data/"))
        evaluate_job.compute = "cpu-cluster"
        # return {
        #     "pipeline_job_trained_model": train_job.outputs.model_output,
        #     "pipeline_job_scored_data": score_job.outputs.score_output,
        #     "pipeline_job_evaluation_report": evaluate_job.outputs.eval_output,
        # }

    pipeline = command_job_in_pipeline(
        Input(path=parent_dir + "/data/"),
        Input(path=parent_dir + "/data/"),
        20,
        1.8,
        "time-based",
    )

    return pipeline



import pytest

from azure.ai.ml.entities import ResourceSettings


@pytest.mark.unittest
@pytest.mark.production_experiences_test
class TestResourceRequirementsSettings:
    def test_resource_requirements_settings_from_object(self) -> None:
        rest_resource_settings = ResourceSettings(cpu="1", memory="1.0Gi", gpu="1")
        resource_settings = ResourceSettings._from_rest_object(settings=rest_resource_settings)
        assert resource_settings.cpu == rest_resource_settings.cpu
        assert resource_settings.memory == rest_resource_settings.memory
        assert resource_settings.gpu == rest_resource_settings.gpu

    def test_resource_requirements_settings_to_objects(self) -> None:
        resource_settings = ResourceSettings(cpu="1", memory="1.0Gi", gpu="1")
        rest_resource_settings = resource_settings._to_rest_object()
        assert resource_settings.cpu == rest_resource_settings.cpu
        assert resource_settings.memory == rest_resource_settings.memory
        assert resource_settings.gpu == rest_resource_settings.gpu



from unittest.mock import DEFAULT, Mock

import pytest
from marshmallow import ValidationError
from pytest_mock import MockFixture

from azure.ai.ml._scope_dependent_operations import OperationScope
from azure.ai.ml.entities import FeatureStore, Workspace
from azure.ai.ml.entities._feature_store._constants import (
    OFFLINE_MATERIALIZATION_STORE_TYPE,
    OFFLINE_STORE_CONNECTION_NAME,
    ONLINE_MATERIALIZATION_STORE_TYPE,
    ONLINE_STORE_CONNECTION_NAME,
)
from azure.ai.ml.entities._feature_store.materialization_store import MaterializationStore
from azure.ai.ml.entities._workspace.feature_store_settings import FeatureStoreSettings
from azure.ai.ml.operations._feature_store_operations import FeatureStoreOperations
from azure.core.polling import LROPoller

MOCK_MATERIALIZATION_STORE_TARGET = "/subscriptions/sub/resourceGroups/rg/providers/Microsoft.Storage/storageAccounts/test_storage/blobServices/default/containers/offlinestore"
MOCK_MATERIALIZATION_STORE_TARGET_INCOMPLETE = (
    "/subscriptions/sub/resourceGroups/rg/providers/Microsoft.Storage/storageAccounts/test_storage"
)


@pytest.fixture
def mock_credential() -> Mock:
    yield Mock()


@pytest.fixture
def mock_feature_store_operation(
    mock_workspace_scope: OperationScope,
    mock_aml_services_2023_08_01_preview: Mock,
    mock_machinelearning_client: Mock,
    mock_credential: Mock,
) -> FeatureStoreOperations:
    yield FeatureStoreOperations(
        operation_scope=mock_workspace_scope,
        service_client=mock_aml_services_2023_08_01_preview,
        all_operations=mock_machinelearning_client._operation_container,
        credentials=mock_credential,
    )


@pytest.mark.unittest
@pytest.mark.data_experiences_test
class TestFeatureStoreOperation:
    @pytest.mark.parametrize("arg", ["resource_group", "subscription", "other_rand_str"])
    def test_list(self, arg: str, mock_feature_store_operation: FeatureStoreOperations) -> None:
        mock_feature_store_operation.list(scope=arg)
        if arg == "subscription":
            mock_feature_store_operation._operation.list_by_subscription.assert_called_once()
        else:
            mock_feature_store_operation._operation.list_by_resource_group.assert_called_once()

    def test_get(self, mock_feature_store_operation: FeatureStoreOperations) -> None:
        def outgoing_get_call(rg, name):
            from azure.ai.ml.entities._workspace.feature_store_settings import FeatureStoreSettings

            ws = Workspace(name=name, kind="featurestore")
            ws._feature_store_settings = FeatureStoreSettings(
                offline_store_connection_name=OFFLINE_STORE_CONNECTION_NAME,
                online_store_connection_name=ONLINE_STORE_CONNECTION_NAME,
            )
            return ws._to_rest_object()

        mock_feature_store_operation._operation.get.side_effect = outgoing_get_call
        featurestore = mock_feature_store_operation.get(name="random_name", resource_group="rg")
        mock_feature_store_operation._operation.get.assert_called_once()

        assert featurestore._kind == "featurestore"
        assert featurestore._feature_store_settings.offline_store_connection_name == OFFLINE_STORE_CONNECTION_NAME
        assert featurestore._feature_store_settings.online_store_connection_name == ONLINE_STORE_CONNECTION_NAME

    def test_begin_create(
        self,
        mock_feature_store_operation: FeatureStoreOperations,
        mocker: MockFixture,
    ):
        from azure.ai.ml.entities._credentials import ManagedIdentityConfiguration

        IDENTITY_RESOURCE_ID = "/subscriptions/b17253fa-f327-42d6-9686-f3e553e24763/resourcegroups/test_rg/providers/Microsoft.ManagedIdentity/userAssignedIdentities/new_identity"

        def outgoing_get_call(rg, name):
            return Workspace(name=name, kind="featurestore")._to_rest_object()

        mocker.patch("azure.ai.ml.operations._feature_store_operations.FeatureStoreOperations.get", return_value=None)
        mocker.patch(
            "azure.ai.ml.operations._feature_store_operations.FeatureStoreOperations._populate_arm_parameters",
            return_value=({}, {}, {}),
        )
        mocker.patch(
            "azure.ai.ml.operations._workspace_operations_base.WorkspaceOperationsBase.begin_create",
            return_value=None,
        )
        mocker.patch("azure.ai.ml._arm_deployments.ArmDeploymentExecutor.deploy_resource", return_value=LROPoller)

        # create
        mock_feature_store_operation._operation.get.side_effect = Exception()
        mock_feature_store_operation.begin_create(feature_store=FeatureStore(name="name"))
        super(type(mock_feature_store_operation), mock_feature_store_operation).begin_create.assert_called_once()

        # create, no materialization identity is allowed
        super(type(mock_feature_store_operation), mock_feature_store_operation).begin_create.reset_mock()
        mock_feature_store_operation.begin_create(
            feature_store=FeatureStore(
                name="name",
                offline_store=MaterializationStore(
                    type=OFFLINE_MATERIALIZATION_STORE_TYPE, target=MOCK_MATERIALIZATION_STORE_TARGET
                ),
            )
        )
        super(type(mock_feature_store_operation), mock_feature_store_operation).begin_create.assert_called_once()
        call_kwargs = super(
            type(mock_feature_store_operation), mock_feature_store_operation
        ).begin_create.call_args.kwargs
        # remove this condition check when test env python version >= 3.8
        if isinstance(call_kwargs, dict):
            feature_store = call_kwargs["workspace"]
            feature_store_settings = feature_store._feature_store_settings
            offline_store_connection_name = feature_store_settings.offline_store_connection_name
            online_store_connection_name = feature_store_settings.online_store_connection_name
            assert offline_store_connection_name.startswith(OFFLINE_STORE_CONNECTION_NAME + "-")
            assert online_store_connection_name == None

        super(type(mock_feature_store_operation), mock_feature_store_operation).begin_create.reset_mock()
        mock_feature_store_operation.begin_create(
            feature_store=FeatureStore(
                name="name",
                online_store=MaterializationStore(
                    type=ONLINE_MATERIALIZATION_STORE_TYPE, target=MOCK_MATERIALIZATION_STORE_TARGET
                ),
            )
        )
        super(type(mock_feature_store_operation), mock_feature_store_operation).begin_create.assert_called_once()
        call_kwargs = super(
            type(mock_feature_store_operation), mock_feature_store_operation
        ).begin_create.call_args.kwargs
        # remove this condition check when test env python version >= 3.8
        if isinstance(call_kwargs, dict):
            feature_store = call_kwargs["workspace"]
            feature_store_settings = feature_store._feature_store_settings
            offline_store_connection_name = feature_store_settings.offline_store_connection_name
            online_store_connection_name = feature_store_settings.online_store_connection_name
            assert offline_store_connection_name.startswith(OFFLINE_STORE_CONNECTION_NAME + "-")
            assert online_store_connection_name.startswith(ONLINE_STORE_CONNECTION_NAME + "-")

        # create, with materialization identity
        super(type(mock_feature_store_operation), mock_feature_store_operation).begin_create.reset_mock()
        mock_feature_store_operation.begin_create(
            feature_store=FeatureStore(
                name="name",
                online_store=MaterializationStore(
                    type=ONLINE_MATERIALIZATION_STORE_TYPE, target=MOCK_MATERIALIZATION_STORE_TARGET
                ),
                offline_store=MaterializationStore(
                    type=OFFLINE_MATERIALIZATION_STORE_TYPE, target=MOCK_MATERIALIZATION_STORE_TARGET
                ),
                materialization_identity=ManagedIdentityConfiguration(resource_id=IDENTITY_RESOURCE_ID),
            )
        )
        super(type(mock_feature_store_operation), mock_feature_store_operation).begin_create.assert_called_once()
        call_kwargs = super(
            type(mock_feature_store_operation), mock_feature_store_operation
        ).begin_create.call_args.kwargs
        # remove this condition check when test env python version >= 3.8
        if isinstance(call_kwargs, dict):
            feature_store = call_kwargs["workspace"]
            feature_store_settings = feature_store._feature_store_settings
            offline_store_connection_name = feature_store_settings.offline_store_connection_name
            online_store_connection_name = feature_store_settings.online_store_connection_name
            assert offline_store_connection_name.startswith(OFFLINE_STORE_CONNECTION_NAME + "-")
            assert online_store_connection_name.startswith(ONLINE_STORE_CONNECTION_NAME + "-")

        # create, with name only
        super(type(mock_feature_store_operation), mock_feature_store_operation).begin_create.reset_mock()
        mock_feature_store_operation.begin_create(feature_store=FeatureStore(name="name"))
        super(type(mock_feature_store_operation), mock_feature_store_operation).begin_create.assert_called_once()
        call_kwargs = super(
            type(mock_feature_store_operation), mock_feature_store_operation
        ).begin_create.call_args.kwargs
        # remove this condition check when test env python version >= 3.8
        if isinstance(call_kwargs, dict):
            feature_store = call_kwargs["workspace"]
            feature_store_settings = feature_store._feature_store_settings
            offline_store_connection_name = feature_store_settings.offline_store_connection_name
            online_store_connection_name = feature_store_settings.online_store_connection_name
            assert offline_store_connection_name.startswith(OFFLINE_STORE_CONNECTION_NAME + "-")
            assert online_store_connection_name == None

        # double create call
        mock_feature_store_operation._operation.get.side_effect = outgoing_get_call
        mock_feature_store_operation._operation.begin_update.side_effect = None
        mock_feature_store_operation.begin_create(feature_store=FeatureStore(name="name"))
        mock_feature_store_operation._operation.begin_update.assert_called()

    def test_update(self, mock_feature_store_operation: FeatureStoreOperations) -> None:
        fs = FeatureStore(
            name="name",
            description="description",
        )

        def outgoing_get_call(rg, name):
            from azure.ai.ml.entities._workspace.feature_store_settings import FeatureStoreSettings

            ws = Workspace(name=name, kind="featurestore")
            ws._feature_store_settings = FeatureStoreSettings()
            return ws._to_rest_object()

        def outgoing_call(rg, name, params, polling, cls):
            assert rg == "test_resource_group"
            assert name == "name"
            assert params.description == "description"
            assert polling is True
            assert callable(cls)
            return DEFAULT

        mock_feature_store_operation._operation.get.side_effect = outgoing_get_call
        mock_feature_store_operation._operation.begin_update.side_effect = outgoing_call
        mock_feature_store_operation.begin_update(fs, update_dependent_resources=True)
        mock_feature_store_operation._operation.begin_update.assert_called()

        with pytest.raises(ValidationError):
            mock_feature_store_operation.begin_update(
                feature_store=FeatureStore(
                    name="name",
                    offline_store=MaterializationStore(
                        type=ONLINE_MATERIALIZATION_STORE_TYPE, target=MOCK_MATERIALIZATION_STORE_TARGET
                    ),
                )
            )
        with pytest.raises(ValidationError):
            mock_feature_store_operation.begin_update(
                feature_store=FeatureStore(
                    name="name",
                    offline_store=MaterializationStore(
                        type=OFFLINE_MATERIALIZATION_STORE_TYPE, target=MOCK_MATERIALIZATION_STORE_TARGET_INCOMPLETE
                    ),
                )
            )
        with pytest.raises(ValidationError):
            mock_feature_store_operation.begin_update(
                feature_store=FeatureStore(
                    name="name",
                    offline_store=MaterializationStore(
                        type=OFFLINE_MATERIALIZATION_STORE_TYPE, target=MOCK_MATERIALIZATION_STORE_TARGET
                    ),
                )
            )
        with pytest.raises(ValidationError):
            mock_feature_store_operation.begin_update(
                feature_store=FeatureStore(
                    name="name",
                    online_store=MaterializationStore(
                        type=ONLINE_MATERIALIZATION_STORE_TYPE, target=MOCK_MATERIALIZATION_STORE_TARGET
                    ),
                )
            )

    def test_update_with_role_assignments(
        self, mock_feature_store_operation: FeatureStoreOperations, mocker: MockFixture
    ) -> None:
        from azure.ai.ml.entities._credentials import ManagedIdentityConfiguration
        from azure.ai.ml.entities._feature_store.materialization_store import MaterializationStore

        EXISTING_IDENTITY_RESOURCE_ID = "/subscriptions/b17253fa-f327-42d6-9686-f3e553e24763/resourcegroups/test_rg/providers/Microsoft.ManagedIdentity/userAssignedIdentities/exist_identity"
        EXISTING_OFFLINE_STORE_RESOURCE_ID = "/subscriptions/b17253fa-f327-42d6-9686-f3e553e24763/resourceGroups/test_rg/providers/Microsoft.Storage/storageAccounts/test_storage/blobServices/default/containers/exist"
        EXISTING_ONLINE_STORE_RESOURCE_ID = "/subscriptions/b17253fa-f327-42d6-9686-f3e553e24763/resourceGroups/mdctest/providers/Microsoft.Cache/Redis/exist"
        OFFLINE_STORE_CONNECTION_NAME = "offlineStoreConnectionName"
        ONLINE_STORE_CONNECTION_NAME = "onlineStoreConnectionName"
        NEW_IDENTITY_RESOURCE_ID = "/subscriptions/b17253fa-f327-42d6-9686-f3e553e24763/resourcegroups/test_rg/providers/Microsoft.ManagedIdentity/userAssignedIdentities/new_identity"
        NEW_OFFLINE_STORE_RESOURCE_ID = "/subscriptions/b17253fa-f327-42d6-9686-f3e553e24763/resourceGroups/test_rg/providers/Microsoft.Storage/storageAccounts/test_storage/blobServices/default/containers/new"
        NEW_ONLINE_STORE_RESOURCE_ID = "/subscriptions/b17253fa-f327-42d6-9686-f3e553e24763/resourceGroups/mdctest/providers/Microsoft.Cache/Redis/new"

        mocker.patch(
            "azure.ai.ml.operations._workspace_operations_base.WorkspaceOperationsBase.begin_update",
            return_value=None,
        )
        mocker.patch(
            "azure.ai.ml.operations._workspace_connections_operations.WorkspaceConnectionsOperations.get",
            return_value=None,
        )

        def outgoing_get_call(rg, name):
            from azure.ai.ml._utils.utils import camel_to_snake
            from azure.ai.ml.constants import ManagedServiceIdentityType
            from azure.ai.ml.entities import IdentityConfiguration, ManagedIdentityConfiguration
            from azure.ai.ml.entities._workspace.feature_store_settings import FeatureStoreSettings

            ws = Workspace(
                name=name,
                kind="featurestore",
                identity=IdentityConfiguration(
                    type=camel_to_snake(ManagedServiceIdentityType.USER_ASSIGNED),
                    user_assigned_identities=[
                        ManagedIdentityConfiguration(resource_id=EXISTING_IDENTITY_RESOURCE_ID),
                    ],
                ),
            )
            ws._feature_store_settings = FeatureStoreSettings(
                offline_store_connection_name=f"{OFFLINE_STORE_CONNECTION_NAME}-existing",
                online_store_connection_name=f"{ONLINE_STORE_CONNECTION_NAME}-existing",
            )
            return ws._to_rest_object()

        def outgoing_workspace_connection_call(resource_group_name, workspace_name, connection_name):
            if connection_name.startswith(OFFLINE_STORE_CONNECTION_NAME):
                from azure.ai.ml._restclient.v2023_04_01_preview.models import (
                    WorkspaceConnectionPropertiesV2,
                    WorkspaceConnectionPropertiesV2BasicResource,
                )

                resource = WorkspaceConnectionPropertiesV2BasicResource(
                    properties=WorkspaceConnectionPropertiesV2(
                        category=OFFLINE_MATERIALIZATION_STORE_TYPE, target=EXISTING_OFFLINE_STORE_RESOURCE_ID
                    )
                )
                return resource
            elif connection_name.startswith(ONLINE_STORE_CONNECTION_NAME):
                from azure.ai.ml._restclient.v2023_04_01_preview.models import (
                    WorkspaceConnectionPropertiesV2,
                    WorkspaceConnectionPropertiesV2BasicResource,
                )

                resource = WorkspaceConnectionPropertiesV2BasicResource(
                    properties=WorkspaceConnectionPropertiesV2(
                        category=ONLINE_MATERIALIZATION_STORE_TYPE, target=EXISTING_ONLINE_STORE_RESOURCE_ID
                    )
                )
                return resource
            else:
                raise Exception("Invalid exception")

        def perform_tests(testcases, mock_feature_store_operation):
            for testcase in testcases:
                mock_feature_store_operation.begin_update(testcase[0], update_dependent_resources=True)
                super(
                    type(mock_feature_store_operation), mock_feature_store_operation
                ).begin_update.assert_called_once()
                call_kwargs = super(
                    type(mock_feature_store_operation), mock_feature_store_operation
                ).begin_update.call_args.kwargs

                # remove this condition check when test env python version >= 3.8
                if isinstance(call_kwargs, dict):
                    feature_store_settings = call_kwargs["feature_store_settings"]
                    offline_store_connection_name = feature_store_settings.offline_store_connection_name
                    online_store_connection_name = feature_store_settings.online_store_connection_name
                    assert call_kwargs["grant_materialization_permissions"] == True
                    assert call_kwargs["update_workspace_role_assignment"] == testcase[1][0]
                    assert call_kwargs["update_offline_store_role_assignment"] == testcase[1][1]
                    assert call_kwargs["update_online_store_role_assignment"] == testcase[1][2]
                    assert call_kwargs["materialization_identity_id"] == testcase[1][3]
                    assert call_kwargs["offline_store_target"] == testcase[1][4]
                    assert call_kwargs["online_store_target"] == testcase[1][5]
                    assert (f"{OFFLINE_STORE_CONNECTION_NAME}-existing" == offline_store_connection_name) == testcase[
                        1
                    ][6]
                    assert (f"{ONLINE_STORE_CONNECTION_NAME}-existing" == online_store_connection_name) == testcase[1][
                        7
                    ]

                print(f"passed test case: {testcase}")
                super(type(mock_feature_store_operation), mock_feature_store_operation).begin_update.reset_mock()

        mock_feature_store_operation._operation.get.side_effect = outgoing_get_call

        # existing stores
        mock_feature_store_operation._workspace_connection_operation.get.side_effect = (
            outgoing_workspace_connection_call
        )

        testcases1 = [
            (FeatureStore(name="name", description="description"), [False, False, False, None, None, None, True, True]),
            (
                FeatureStore(
                    name="name",
                    description="description",
                    materialization_identity=ManagedIdentityConfiguration(resource_id=NEW_IDENTITY_RESOURCE_ID),
                ),
                [
                    True,
                    True,
                    True,
                    NEW_IDENTITY_RESOURCE_ID,
                    EXISTING_OFFLINE_STORE_RESOURCE_ID,
                    EXISTING_ONLINE_STORE_RESOURCE_ID,
                    True,
                    True,
                ],
            ),
            (
                FeatureStore(
                    name="name",
                    description="description",
                    offline_store=MaterializationStore(
                        type=OFFLINE_MATERIALIZATION_STORE_TYPE, target=NEW_OFFLINE_STORE_RESOURCE_ID
                    ),
                ),
                [False, True, False, EXISTING_IDENTITY_RESOURCE_ID, NEW_OFFLINE_STORE_RESOURCE_ID, None, False, True],
            ),
            (
                FeatureStore(
                    name="name",
                    description="description",
                    online_store=MaterializationStore(
                        type=ONLINE_MATERIALIZATION_STORE_TYPE, target=NEW_ONLINE_STORE_RESOURCE_ID
                    ),
                ),
                [False, False, True, EXISTING_IDENTITY_RESOURCE_ID, None, NEW_ONLINE_STORE_RESOURCE_ID, True, False],
            ),
            (
                FeatureStore(
                    name="name",
                    description="description",
                    materialization_identity=ManagedIdentityConfiguration(resource_id=NEW_IDENTITY_RESOURCE_ID),
                    offline_store=MaterializationStore(
                        type=OFFLINE_MATERIALIZATION_STORE_TYPE, target=NEW_OFFLINE_STORE_RESOURCE_ID
                    ),
                    online_store=MaterializationStore(
                        type=ONLINE_MATERIALIZATION_STORE_TYPE, target=NEW_ONLINE_STORE_RESOURCE_ID
                    ),
                ),
                [
                    True,
                    True,
                    True,
                    NEW_IDENTITY_RESOURCE_ID,
                    NEW_OFFLINE_STORE_RESOURCE_ID,
                    NEW_ONLINE_STORE_RESOURCE_ID,
                    False,
                    False,
                ],
            ),
            (
                FeatureStore(
                    name="name",
                    description="description",
                    materialization_identity=ManagedIdentityConfiguration(resource_id=NEW_IDENTITY_RESOURCE_ID),
                    offline_store=MaterializationStore(
                        type=OFFLINE_MATERIALIZATION_STORE_TYPE, target=EXISTING_OFFLINE_STORE_RESOURCE_ID
                    ),
                    online_store=MaterializationStore(
                        type=ONLINE_MATERIALIZATION_STORE_TYPE, target=EXISTING_ONLINE_STORE_RESOURCE_ID
                    ),
                ),
                [
                    True,
                    True,
                    True,
                    NEW_IDENTITY_RESOURCE_ID,
                    EXISTING_OFFLINE_STORE_RESOURCE_ID,
                    EXISTING_ONLINE_STORE_RESOURCE_ID,
                    True,
                    True,
                ],
            ),
            (
                FeatureStore(
                    name="name",
                    description="description",
                    materialization_identity=ManagedIdentityConfiguration(resource_id=EXISTING_IDENTITY_RESOURCE_ID),
                    offline_store=MaterializationStore(
                        type=OFFLINE_MATERIALIZATION_STORE_TYPE, target=NEW_OFFLINE_STORE_RESOURCE_ID
                    ),
                    online_store=MaterializationStore(
                        type=ONLINE_MATERIALIZATION_STORE_TYPE, target=NEW_ONLINE_STORE_RESOURCE_ID
                    ),
                ),
                [
                    False,
                    True,
                    True,
                    EXISTING_IDENTITY_RESOURCE_ID,
                    NEW_OFFLINE_STORE_RESOURCE_ID,
                    NEW_ONLINE_STORE_RESOURCE_ID,
                    False,
                    False,
                ],
            ),
            (
                FeatureStore(
                    name="name",
                    description="description",
                    offline_store=MaterializationStore(
                        type=OFFLINE_MATERIALIZATION_STORE_TYPE, target=EXISTING_OFFLINE_STORE_RESOURCE_ID
                    ),
                    online_store=MaterializationStore(
                        type=ONLINE_MATERIALIZATION_STORE_TYPE, target=EXISTING_ONLINE_STORE_RESOURCE_ID
                    ),
                ),
                [False, False, False, None, None, None, True, True],
            ),
        ]

        perform_tests(testcases1, mock_feature_store_operation)

        # non existing stores
        def outgoing_get_call_empty_no_workspace_connection(rg, name):
            from azure.ai.ml._utils.utils import camel_to_snake
            from azure.ai.ml.constants import ManagedServiceIdentityType
            from azure.ai.ml.entities import IdentityConfiguration, ManagedIdentityConfiguration
            from azure.ai.ml.entities._workspace.feature_store_settings import FeatureStoreSettings

            ws = Workspace(
                name=name,
                kind="featurestore",
                identity=IdentityConfiguration(
                    type=camel_to_snake(ManagedServiceIdentityType.USER_ASSIGNED),
                    user_assigned_identities=[
                        ManagedIdentityConfiguration(resource_id=EXISTING_IDENTITY_RESOURCE_ID),
                    ],
                ),
            )
            ws._feature_store_settings = FeatureStoreSettings(
                offline_store_connection_name=None,
                online_store_connection_name=None,
            )
            return ws._to_rest_object()

        mock_feature_store_operation._operation.get.side_effect = outgoing_get_call_empty_no_workspace_connection

        testcases2 = [
            (
                FeatureStore(name="name", description="description"),
                [False, False, False, None, None, None, False, False],
            ),
            (
                FeatureStore(
                    name="name",
                    description="description",
                    materialization_identity=ManagedIdentityConfiguration(resource_id=NEW_IDENTITY_RESOURCE_ID),
                ),
                [True, False, False, NEW_IDENTITY_RESOURCE_ID, None, None, False, False],
            ),
            (
                FeatureStore(
                    name="name",
                    description="description",
                    offline_store=MaterializationStore(
                        type=OFFLINE_MATERIALIZATION_STORE_TYPE, target=NEW_OFFLINE_STORE_RESOURCE_ID
                    ),
                    online_store=MaterializationStore(
                        type=ONLINE_MATERIALIZATION_STORE_TYPE, target=NEW_ONLINE_STORE_RESOURCE_ID
                    ),
                ),
                [
                    False,
                    True,
                    True,
                    EXISTING_IDENTITY_RESOURCE_ID,
                    NEW_OFFLINE_STORE_RESOURCE_ID,
                    NEW_ONLINE_STORE_RESOURCE_ID,
                    False,
                    False,
                ],
            ),
            (
                FeatureStore(
                    name="name",
                    description="description",
                    materialization_identity=ManagedIdentityConfiguration(resource_id=EXISTING_IDENTITY_RESOURCE_ID),
                ),
                [False, False, False, None, None, None, False, False],
            ),
        ]

        perform_tests(testcases2, mock_feature_store_operation)

    def test_begin_provision_network(self, mock_feature_store_operation: FeatureStoreOperations) -> None:
        def outgoing_call(name):
            return "testws"

        mock_feature_store_operation._operation.get.side_effect = outgoing_call
        mock_feature_store_operation.begin_provision_network(feature_store_name="testws")
        mock_feature_store_operation._provision_network_operation.begin_provision_managed_network.assert_called_once()

    def test_delete(self, mock_feature_store_operation: FeatureStoreOperations, mocker: MockFixture) -> None:
        def outgoing_call(rg, name):
            return Workspace(name=name, kind="featurestore")._to_rest_object()

        mock_feature_store_operation._operation.get.side_effect = outgoing_call
        mocker.patch("azure.ai.ml.operations._workspace_operations_base.delete_resource_by_arm_id", return_value=None)
        mocker.patch(
            "azure.ai.ml.operations._workspace_operations_base.get_generic_arm_resource_by_arm_id", return_value=None
        )
        mock_feature_store_operation.begin_delete("randstr", delete_dependent_resources=True)
        mock_feature_store_operation._operation.begin_delete.assert_called_once()

    def test_delete_non_feature_store_kind(
        self, mock_feature_store_operation: FeatureStoreOperations, mocker: MockFixture
    ) -> None:
        def outgoing_call(rg, name):
            return Workspace(name=name)._to_rest_object()

        mock_feature_store_operation._operation.get.side_effect = outgoing_call
        mocker.patch("azure.ai.ml.operations._workspace_operations_base.delete_resource_by_arm_id", return_value=None)
        with pytest.raises(Exception):
            mock_feature_store_operation.begin_delete("randstr", delete_dependent_resources=True)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

import json
import logging
from typing import Any, Iterable, List, Optional

from marshmallow.exceptions import ValidationError as SchemaValidationError

from azure.ai.ml._exception_helper import log_and_raise_error
from azure.ai.ml._local_endpoints import EndpointStub
from azure.ai.ml._local_endpoints.docker_client import (
    DockerClient,
    get_endpoint_json_from_container,
    get_scoring_uri_from_container,
    get_status_from_container,
)
from azure.ai.ml._utils._endpoint_utils import local_endpoint_polling_wrapper
from azure.ai.ml._utils._http_utils import HttpPipeline
from azure.ai.ml._utils.utils import DockerProxy
from azure.ai.ml.constants._endpoint import EndpointInvokeFields, LocalEndpointConstants
from azure.ai.ml.entities import OnlineEndpoint
from azure.ai.ml.exceptions import InvalidLocalEndpointError, LocalEndpointNotFoundError, ValidationException

docker = DockerProxy()
module_logger = logging.getLogger(__name__)


class _LocalEndpointHelper(object):
    """A helper class to interact with Azure ML endpoints locally.

    Use this helper to manage Azure ML endpoints locally, e.g. create, invoke, show, list, delete.
    """

    def __init__(self, *, requests_pipeline: HttpPipeline):
        self._docker_client = DockerClient()
        self._endpoint_stub = EndpointStub()
        self._requests_pipeline = requests_pipeline

    def create_or_update(self, endpoint: OnlineEndpoint) -> OnlineEndpoint:  # type: ignore
        """Create or update an endpoint locally using Docker.

        :param endpoint: OnlineEndpoint object with information from user yaml.
        :type endpoint: OnlineEndpoint
        """
        try:
            if endpoint is None:
                msg = "The entity provided for local endpoint was null. Please provide valid entity."
                raise InvalidLocalEndpointError(message=msg, no_personal_data_message=msg)

            try:
                self.get(endpoint_name=str(endpoint.name))
                operation_message = "Updating local endpoint"
            except LocalEndpointNotFoundError:
                operation_message = "Creating local endpoint"

            local_endpoint_polling_wrapper(
                func=self._endpoint_stub.create_or_update,
                message=f"{operation_message} ({endpoint.name}) ",
                endpoint=endpoint,
            )
            return self.get(endpoint_name=str(endpoint.name))
        except Exception as ex:  # pylint: disable=W0718
            if isinstance(ex, (ValidationException, SchemaValidationError)):
                log_and_raise_error(ex)
            else:
                raise ex

    def invoke(self, endpoint_name: str, data: dict, deployment_name: Optional[str] = None) -> str:
        """Invoke a local endpoint.

        :param endpoint_name: Name of endpoint to invoke.
        :type endpoint_name: str
        :param data: json data to pass
        :type data: dict
        :param deployment_name: Name of specific deployment to invoke.
        :type deployment_name: (str, optional)
        :return: The text response
        :rtype: str
        """
        # get_scoring_uri will throw user error if there are multiple deployments and no deployment_name is specified
        scoring_uri = self._docker_client.get_scoring_uri(endpoint_name=endpoint_name, deployment_name=deployment_name)
        if scoring_uri:
            headers = {}
            if deployment_name is not None:
                headers[EndpointInvokeFields.MODEL_DEPLOYMENT] = deployment_name
            return str(self._requests_pipeline.post(scoring_uri, json=data, headers=headers).text())
        endpoint_stub = self._endpoint_stub.get(endpoint_name=endpoint_name)
        if endpoint_stub:
            return str(self._endpoint_stub.invoke())
        raise LocalEndpointNotFoundError(endpoint_name=endpoint_name, deployment_name=deployment_name)

    def get(self, endpoint_name: str) -> OnlineEndpoint:
        """Get a local endpoint.

        :param endpoint_name: Name of endpoint.
        :type endpoint_name: str
        :return OnlineEndpoint:
        """
        endpoint = self._endpoint_stub.get(endpoint_name=endpoint_name)
        container = self._docker_client.get_endpoint_container(endpoint_name=endpoint_name, include_stopped=True)
        if endpoint:
            if container:
                return _convert_container_to_endpoint(container=container, endpoint_json=endpoint.dump())
            return endpoint
        if container:
            return _convert_container_to_endpoint(container=container)
        raise LocalEndpointNotFoundError(endpoint_name=endpoint_name)

    def list(self) -> Iterable[OnlineEndpoint]:
        """List all local endpoints.

        :return: An iterable of local endpoints
        :rtype: Iterable[OnlineEndpoint]
        """
        endpoints: List = []
        containers = self._docker_client.list_containers()
        endpoint_stubs = self._endpoint_stub.list()
        # Iterate through all cached endpoint files
        for endpoint_file in endpoint_stubs:
            endpoint_json = json.loads(endpoint_file.read_text())
            container = self._docker_client.get_endpoint_container(
                endpoint_name=endpoint_json.get("name"), include_stopped=True
            )
            # If a deployment is associated with endpoint,
            # override certain endpoint properties with deployment information and remove it from containers list.
            # Otherwise, return endpoint spec.
            if container:
                endpoints.append(_convert_container_to_endpoint(endpoint_json=endpoint_json, container=container))
                containers.remove(container)
            else:
                endpoints.append(
                    OnlineEndpoint._load(
                        data=endpoint_json,
                        params_override=[{"location": LocalEndpointConstants.ENDPOINT_STATE_LOCATION}],
                    )
                )
        # Iterate through any deployments that don't have an explicit local endpoint stub.
        for container in containers:
            endpoints.append(_convert_container_to_endpoint(container=container))
        return endpoints

    def delete(self, name: str) -> None:
        """Delete a local endpoint.

        :param name: Name of endpoint to delete.
        :type name: str
        """
        endpoint_stub = self._endpoint_stub.get(endpoint_name=name)
        if endpoint_stub:
            self._endpoint_stub.delete(endpoint_name=name)
            endpoint_container = self._docker_client.get_endpoint_container(endpoint_name=name)
            if endpoint_container:
                self._docker_client.delete(endpoint_name=name)
        else:
            raise LocalEndpointNotFoundError(endpoint_name=name)


def _convert_container_to_endpoint(
    # Bug Item number: 2885719
    container: "docker.models.containers.Container",  # type: ignore
    endpoint_json: Optional[dict] = None,
) -> OnlineEndpoint:
    """Converts provided Container for local deployment to OnlineEndpoint entity.

    :param container: Container for a local deployment.
    :type container: docker.models.containers.Container
    :param endpoint_json: The endpoint json
    :type endpoint_json: Optional[dict]
    :return: The OnlineEndpoint entity
    :rtype: OnlineEndpoint
    """
    if endpoint_json is None:
        endpoint_json = get_endpoint_json_from_container(container=container)
    provisioning_state = get_status_from_container(container=container)
    if provisioning_state == LocalEndpointConstants.CONTAINER_EXITED:
        return _convert_json_to_endpoint(
            endpoint_json=endpoint_json,
            location=LocalEndpointConstants.ENDPOINT_STATE_LOCATION,
            provisioning_state=LocalEndpointConstants.ENDPOINT_STATE_FAILED,
        )
    scoring_uri = get_scoring_uri_from_container(container=container)
    return _convert_json_to_endpoint(
        endpoint_json=endpoint_json,
        location=LocalEndpointConstants.ENDPOINT_STATE_LOCATION,
        provisioning_state=LocalEndpointConstants.ENDPOINT_STATE_SUCCEEDED,
        scoring_uri=scoring_uri,
    )


def _convert_json_to_endpoint(endpoint_json: Optional[dict], **kwargs: Any) -> OnlineEndpoint:
    """Converts metadata json and kwargs to OnlineEndpoint entity.

    :param endpoint_json: dictionary representation of OnlineEndpoint entity.
    :type endpoint_json: dict
    :return: The OnlineEndpoint entity
    :rtype: OnlineEndpoint
    """
    params_override = []
    for k, v in kwargs.items():
        params_override.append({k: v})
    return OnlineEndpoint._load(data=endpoint_json, params_override=params_override)  # type: ignore
